import tensorflow as tf
import numpy as np

from ..utils import generate_model, generate_data, almost_equal
from xplique.metrics import MeGe
from xplique.attributions import Saliency, GradientInput


def test_best_mege():
    # ensure we get perfect score when the models are the same
    input_shape, nb_labels, nb_samples = ((8, 8, 1), 10, 80)
    x, y = generate_data(input_shape, nb_labels, nb_samples)

    model = generate_model(input_shape, nb_labels)
    learning_algorithm = lambda x_train, y_train, x_test, y_test: model

    for method in [Saliency, GradientInput]:

        metric = MeGe(learning_algorithm, x, y, k_splits=4)
        mege, _ = metric.evaluate(method)

        assert almost_equal(mege, 1.0)



"""
Utility functions to visualize explanations
"""
from .image import plot_attributions, plot_attribution, plot_maco
from .tabular import plot_feature_impact, plot_mean_feature_impact, summary_plot_tabular



import numpy as np

import torch.nn as nn
import tensorflow as tf

import pytest

from xplique.wrappers import TorchWrapper
from xplique.attributions import (Saliency, GradientInput, IntegratedGradients, SmoothGrad, VarGrad,
                                  SquareGrad, Occlusion, Rise, SobolAttributionMethod, Lime, KernelShap,
                                  HsicAttributionMethod)
from xplique.metrics import MuFidelity, Deletion, Insertion, AverageStability
from xplique.metrics.base import ExplainerMetric

from ..utils import generate_data

def map_four_by_four(inp):

    width = inp.shape[0]
    height = inp.shape[1]

    mapping = np.zeros((width,height))
    for i in range(width):
        if i%2 != 0:
            mapping[i] = mapping[i-1]
        else:
            for j in range(height):
                mapping[i][j] = (width/2) * (i//2) + (j//2)

    mapping = tf.cast(mapping, dtype=tf.int32)
    return mapping

def _default_methods_cnn(model):
    return [
        Saliency(model),
        GradientInput(model),
        IntegratedGradients(model),
        SmoothGrad(model),
        SquareGrad(model),
        VarGrad(model),
        Occlusion(model),
        Rise(model),
        SobolAttributionMethod(model),
        HsicAttributionMethod(model),
        Lime(model, nb_samples = 20, map_to_interpret_space=map_four_by_four),
        KernelShap(model, nb_samples = 20, map_to_interpret_space=map_four_by_four),
    ]

def _default_methods_regression(model):
    return [
        Saliency(model, operator="regression"),
        GradientInput(model, operator="regression"),
        IntegratedGradients(model, operator="regression"),
        SmoothGrad(model, operator="regression"),
        SquareGrad(model, operator="regression"),
        VarGrad(model, operator="regression"),
        Occlusion(model, operator="regression", patch_size=1, patch_stride=1),
        Lime(model, operator="regression", nb_samples = 20),
        KernelShap(model, operator="regression", nb_samples = 20),
    ]

def generate_torch_model(input_shape=(32, 32, 3), output_shape=10):
    c_in = input_shape[-1]
    h_in = input_shape[0]
    w_in = input_shape[1]

    model = nn.Sequential()

    model.append(nn.Conv2d(c_in, 4, (2, 2)))
    h_out = h_in - 1
    w_out = w_in -1
    c_out = 4

    model.append(nn.ReLU())
    model.append(nn.MaxPool2d((2, 2)))
    h_out = int((h_out - 2)/2 + 1)
    w_out = int((w_out - 2)/2 + 1)

    model.append(nn.Dropout(0.25))
    model.append(nn.Flatten())
    flatten_size = c_out * h_out * w_out

    model.append(nn.Linear(int(flatten_size) ,output_shape))

    return model

def generate_regression_torch(features_shape, output_shape=1):
    in_size = np.prod(features_shape)
    model = nn.Sequential()
    model.append(nn.Flatten())
    model.append(nn.Linear(in_size, 4))
    model.append(nn.ReLU())
    model.append(nn.Linear(4, 4))
    model.append(nn.ReLU())
    model.append(nn.Linear(4, output_shape))

    return model

def test_assert_eval_mode():
    input_shapes = [(28, 28, 1), (32, 32, 3)]
    nb_labels = 10

    for input_shape in input_shapes:
        model = generate_torch_model(input_shape, nb_labels)
        with pytest.raises(AssertionError):
            wrapped_model = TorchWrapper(model, device='cpu')
        model.eval()
        wrapped_model = TorchWrapper(model, device='cpu')

def test_cnn_wrapper():
    input_shapes = [(28, 28, 1), (32, 32, 3)]
    nb_labels = 10

    for input_shape in input_shapes:
        x, y = generate_data(input_shape, nb_labels, 16)
        model = generate_torch_model(input_shape, nb_labels)
        model.eval()
        wrapped_model = TorchWrapper(model, device='cpu')

        explainers = _default_methods_cnn(wrapped_model)

        for explainer in explainers:
            saliency_maps = explainer.explain(x, y)
            assert x.shape[:3] == saliency_maps.shape[:3]

def test_dense_wrapper():

    features_shape, output_shape, samples = ((10,), 1, 16)
    model = generate_regression_torch(features_shape, output_shape)
    model.eval()
    x, y = generate_data(features_shape, output_shape, samples)
    wrapped_model = TorchWrapper(model, device='cpu', is_channel_first=False)

    explainers = _default_methods_regression(wrapped_model)

    for explainer in explainers:
        try:
            explanations = explainer.explain(x, y)
        except:
            raise AssertionError(
                "Explanation failed for method ", explainer.__class__.__name__)
        assert explanations.shape == [samples, *features_shape]

def test_metric_cnn():
    """Test that wrapped torch model can also leverage the metrics module"""

    input_shape, nb_labels, samples = ((16, 16, 3), 5, 8)
    x, y = generate_data(input_shape, nb_labels, samples)
    model = generate_torch_model(input_shape, nb_labels)
    model.eval()
    wrapped_model = TorchWrapper(model, device='cpu')

    explainers = _default_methods_cnn(wrapped_model)

    metrics = [
        Deletion(wrapped_model, x, y, steps=3),
        Insertion(wrapped_model, x, y, steps=3),
        MuFidelity(wrapped_model, x, y, nb_samples=3),
        AverageStability(wrapped_model, x, y, nb_samples=3)
    ]

    for explainer in explainers:
        explanations = explainer(x, y)
        for metric in metrics:
            assert hasattr(metric, 'evaluate')
            if isinstance(metric, ExplainerMetric):
                score = metric(explainer)
            else:
                assert hasattr(metric, 'inference_function')
                assert hasattr(metric, 'batch_inference_function')
                score = metric(explanations)
            assert type(score) in [np.float32, np.float64, float]

def test_metric_dense():
    """Test that wrapped torch model can also leverage the metrics module"""

    input_shape, nb_labels, samples = ((16, 16, 3), 5, 8)
    x, y = generate_data(input_shape, nb_labels, samples)
    model = generate_regression_torch(input_shape, nb_labels)
    model.eval()
    wrapped_model = TorchWrapper(model, device='cpu')

    explainers = _default_methods_regression(wrapped_model)

    metrics = [
        Deletion(wrapped_model, x, y, steps=3, activation="sigmoid"),
        Insertion(wrapped_model, x, y, steps=3, activation="softmax"),
        MuFidelity(wrapped_model, x, y, nb_samples=3),
        AverageStability(wrapped_model, x, y, nb_samples=3)
    ]

    for explainer in explainers:
        explanations = explainer(x, y)
        for metric in metrics:
            assert hasattr(metric, 'evaluate')
            if isinstance(metric, ExplainerMetric):
                score = metric(explainer)
            else:
                assert hasattr(metric, 'inference_function')
                assert hasattr(metric, 'batch_inference_function')
                score = metric(explanations)
            assert type(score) in [np.float32, np.float64, float]



"""
Test for the wrapper
Also serve as a generic test
"""

import numpy as np
import tensorflow as tf

from xplique.attributions import KernelShap, Lime, Occlusion
from ..utils import generate_dataset, generate_linear_model, almost_equal


def test_wrapper():
    """
    Test the wrapping of a sklearn model, by comparing it to a keras model.
    Both are linear regression model with really close weights.
    We expect the explanations on those models to be close as well.
    """
    nb_samples = 100
    epsilon = 1.e-4 * nb_samples
    dataset, _, features_coef = generate_dataset(nb_samples)

    libraries = [
        "sklearn",
        "keras",
    ]

    models = {
        lib: generate_linear_model(features_coef, lib)
        for lib in libraries
    }

    # we want both models to be close enough
    y_preds = [np.array(model(dataset)).reshape((nb_samples,)) for model in models.values()]
    assert almost_equal(y_preds[0], y_preds[1], epsilon)

    # Define explainers
    model_explainers = {}
    for library, model in models.items():
        model_explainers[library] = {
            "KernelShap": KernelShap(
                model,
                nb_samples=200,  # 2000
                ref_value=0.0,
                batch_size=nb_samples
            ),
            "Lime": Lime(
                model,
                nb_samples=2000,  # 2000
                ref_value=0.0,
                distance_mode="cosine",
                kernel_width=1.0,
                batch_size=nb_samples
            ),
            "Occlusion": Occlusion(
                model,
                patch_size=1,
                patch_stride=1,
                occlusion_value=0.0,
                batch_size=nb_samples
            ),
        }

    # cast a subset to tf format
    inputs_tf = tf.cast(dataset, tf.float32)
    targets_tf = tf.ones((len(dataset), 1))

    explanations = {}

    # compute explanations for all methods
    for library, explainers in model_explainers.items():
        explanations[library] = {}
        for method, explainer in explainers.items():
            # Lime is not stable compared, a seed is needed
            tf.random.set_seed(0)
            # compute explanation for a given method
            explanations[library][method] = explainer(inputs_tf, targets_tf)

    for method in explanations[libraries[0]].keys():
        expl0 = explanations[libraries[0]][method]
        expl1 = explanations[libraries[1]][method]
        assert almost_equal(expl0, expl1, epsilon)



"""
HSIC estimator
"""
# pylint: disable=C0103

from abc import ABC, abstractmethod
from functools import partial

import tensorflow as tf
import numpy as np

from .kernels import Kernel


class HsicEstimator(ABC):
    """
    Base class for HSIC estimator.
    """

    def __init__(self, output_kernel="rbf"):
        self.output_kernel = output_kernel
        assert output_kernel in [
            "rbf"
        ], "Only 'rbf' output kernel is supported for now."

    @staticmethod
    def masks_dim(masks):
        """
        Deduce the number of dimensions using the sampling masks.

        Parameters
        ----------
        masks
          Low resolution masks (before upsampling) used, one for each output.

        Returns
        -------
        nb_dim
          The number of dimensions under study according to the masks.
        """
        nb_dim = np.prod(masks.shape[1:])
        return nb_dim

    @staticmethod
    def post_process(score, masks):
        """
        Post processing ops on the indices before sending them back. Makes sure the data
        format and shape is correct.

        Parameters
        ----------
        score
          Total order HSIC scores, one for each dimensions.
        masks
            Low resolution masks (before upsampling) used, one for each output.

        Returns
        -------
        score
          HSIC scores after post processing.
        """
        score = np.array(score, np.float32)
        return np.transpose(score.reshape(masks.shape[1:]), axes=(1, 0, 2))

    @abstractmethod
    def input_kernel_func(self, X, Y):
        """
        Kernel function for the input.

        Parameters
        ----------
        X
            Samples of input variable
        Y
            Samples of output variable

        Returns
        -------
        Kernel matrix
        """
        raise NotImplementedError()

    @abstractmethod
    def output_kernel_func(self, X, Y):
        """
        kernel function for the output

        Parameters
        ----------
        X
            Samples of input variable
        Y
            Samples of output variable

        Returns
        -------
        Kernel matrix
        """
        raise NotImplementedError()

    @tf.function
    def estimator(self, masks, L, nb_dim, nb_design):
        """
        tf operations related to the estimator for performances

        Parameters
        ----------
        masks
            binary masks, each dimension corresponding to an image patch
        L
            output samples kernel Gram matrix
        nb_dim
            number of input variables to consider
        nb_design
            number of points used to estimate HSIC

        Returns
        -------
        HSIC estimates
            Raw HSIC estimates in tensorflow
        """
        X = tf.transpose(masks)

        X1 = tf.reshape(X, (nb_dim, 1, nb_design, 1))
        X2 = tf.transpose(X1, [0, 1, 3, 2])
        K = self.input_kernel_func(X1, X2)
        K = tf.math.reduce_prod(1 + K, axis=1)

        H = tf.eye(nb_design) - tf.ones((nb_design, nb_design)) / nb_design
        HK = tf.einsum("jk,ikl->ijl", H, K)
        HL = tf.einsum("jk,kl->jl", H, L)

        Kc = tf.einsum("ijk,kl->ijl", HK, H)
        Lc = tf.einsum("jk,kl->jl", HL, H)

        score = tf.math.reduce_sum(Kc * tf.transpose(Lc), axis=[1, 2]) / nb_design
        return score

    def __call__(self, masks, outputs, nb_design):
        """
        Compute the test statistic using a self.output_kernel_func kernel for the output
        and self.input_kernel_func for the input, to be defined in child classes
        for the input.

        Parameters
        ----------
        masks
            binary masks, each dimension corresponding to an image patch
        outputs
            samples of the output variable
        nb_design
            number of points used to estimate HSIC

        Returns
        -------
        HSIC estimates
            Array with HSIC estimates for each patch
        """
        nb_dim = self.masks_dim(masks)

        Y = tf.cast(outputs, tf.float32)
        Y = tf.reshape(Y, (nb_design, 1))
        L = self.output_kernel_func(Y, tf.transpose(Y))

        score = self.estimator(masks, L, nb_dim, nb_design)

        return self.post_process(score, masks)


class BinaryEstimator(HsicEstimator):
    """
    Estimator based on the Dirac kernel for the input
    """

    def input_kernel_func(self, X, Y):
        return Kernel.from_string("binary")(X, Y)

    def output_kernel_func(self, X, Y):
        width_y = np.percentile(Y, 50.0).astype(np.float32)
        kernel_func = partial(Kernel.from_string(self.output_kernel), width=width_y)
        return kernel_func(X, Y)


class RbfEstimator(HsicEstimator):
    """
    Estimator based on the RBF kernel for the input
    """

    def input_kernel_func(self, X, Y):
        width_x = 0.5
        kernel_func = partial(Kernel.from_string("rbf"), width=width_x)
        return kernel_func(X, Y)

    def output_kernel_func(self, X, Y):
        width_y = np.percentile(Y, 50.0).astype(np.float32)
        kernel_func = partial(Kernel.from_string(self.output_kernel), width=width_y)
        return kernel_func(X, Y)


class SobolevEstimator(HsicEstimator):
    """
    Estimator based on the Sobolev kernel for the input
    """

    def input_kernel_func(self, X, Y):
        return Kernel.from_string("sobolev")(X, Y)

    def output_kernel_func(self, X, Y):
        width_y = np.percentile(Y, 50.0).astype(np.float32)
        kernel_func = partial(Kernel.from_string(self.output_kernel), width=width_y)
        return kernel_func(X, Y)



"""
Attributions methods availables
"""

from .saliency import Saliency
from .gradient_input import GradientInput
from .grad_cam import GradCAM
from .integrated_gradients import IntegratedGradients
from .smoothgrad import SmoothGrad
from .vargrad import VarGrad
from .square_grad import SquareGrad
from .occlusion import Occlusion
from .rise import Rise
from .guided_backpropagation import GuidedBackprop
from .deconvnet import DeconvNet
from .grad_cam_pp import GradCAMPP
from .lime import Lime
from .kernel_shap import KernelShap
from .object_detector import BoundingBoxesExplainer
from .global_sensitivity_analysis import SobolAttributionMethod, HsicAttributionMethod
from . import global_sensitivity_analysis



import tensorflow as tf
import numpy as np

from xplique.features_visualizations import maco_image_parametrization, init_maco_buffer, maco, Objective
from ..utils import almost_equal


def dummy_model():
    model = tf.keras.Sequential([
        tf.keras.layers.Input((8, 8, 3)),
        tf.keras.layers.Conv2D(4, (3, 3)),
        tf.keras.layers.Flatten(),
        tf.keras.layers.Dense(10)
    ])
    model.compile()
    return model


def test_init_maco_buffer():
    """ Ensure we can init the magnitude and phase for any size """
    img_size_to_magnitude_size = {
        (8, 8):(8, 5),
        (16, 24):(16, 13),
        (32, 32):(32, 17),
        (128, 128):(128, 65),
        (256, 512):(256, 257),
        (1024, 2048):(1024, 1025),
    }

    for img_size, spectrum_size in img_size_to_magnitude_size.items():
        magnitude, phase = init_maco_buffer(img_size)
        assert magnitude.shape[1:] == spectrum_size
        assert phase.shape[1:] == spectrum_size


def test_maco_image_param():
    """ Ensure we can reconstruct an image from magnitude and phase """
    img_size_to_magnitude_size = {
        (8, 8):(8, 5),
        (16, 24):(16, 13),
        (32, 32):(32, 17),
        (128, 128):(128, 65),
        (256, 512):(256, 257),
        (1024, 2048):(1024, 1025),
    }

    for img_size, spectrum_size in img_size_to_magnitude_size.items():
        magnitude, phase = init_maco_buffer(img_size)
        img = maco_image_parametrization(magnitude, phase, (0, 1))
        assert img.shape[:-1] == img_size


def test_maco():
    """ Ensure the optimization process is returning a valid image """
    model = dummy_model()

    objectives = [
        Objective.neuron(model, -1, 0),
        Objective.direction(model, -1, tf.one_hot(1, 10)),
        Objective.channel(model, -3, 0),
    ]

    for objective in objectives:
        img, transparency = maco(objective, nb_steps=10, custom_shape=(8, 8), values_range=(-127, 127))
        assert img.shape == (8, 8, 3)
        assert transparency.shape == (8, 8, 3)
        assert np.min(img) >= -127
        assert np.max(img) <= 127



"""
Module related to Kernel SHAP method
"""

import tensorflow as tf
import numpy as np
from sklearn import linear_model

from .lime import Lime
from ..commons import Tasks
from ..types import Callable, Union, Optional, OperatorSignature

class KernelShap(Lime):
    """
    By setting appropriately the perturbation function, the similarity kernel and the interpretable
    model in the LIME framework we can theoretically obtain the Shapley Values more efficiently.
    Therefore, KernelShap is a method based on LIME with specific attributes.

    More information regarding this method and proof of equivalence can be found in the
    original paper here:
    https://arxiv.org/abs/1705.07874

    Parameters
    ----------
    model
        The model from which we want to obtain explanations.
    batch_size
        Number of perturbed samples to process at once, mandatory when nb_samples is huge.
        Notice, it is different compare to WhiteBox explainers which batch the inputs.
        Here inputs are process one by one.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].
    map_to_interpret_space
        Function which group features of an input corresponding to the same interpretable
        feature (e.g super-pixel).
        It allows to transpose from (resp. to) the original input space to (resp. from)
        the interpretable space.
    nb_samples
        The number of perturbed samples you want to generate for each input sample.
        Default to 800.
    ref_value
        It defines reference value which replaces each feature when the corresponding
        interpretable feature is set to 0.
        It should be provided as: a ndarray of shape (1) if there is no channels in your input
        and (C,) otherwise.
        The default ref value is set to (0.5,0.5,0.5) for inputs with 3 channels (corresponding
        to a grey pixel when inputs are normalized by 255) and to 0 otherwise.
    """
    def __init__(self,
                 model: Callable,
                 batch_size: int = 64,
                 operator: Optional[Union[Tasks, str, OperatorSignature]] = None,
                 map_to_interpret_space: Optional[Callable] = None,
                 nb_samples: int = 800,
                 ref_value: Optional[np.ndarray] = None):

        Lime.__init__(
            self,
            model,
            batch_size,
            operator,
            interpretable_model = linear_model.LinearRegression(),
            similarity_kernel = KernelShap._kernel_shap_similarity_kernel,
            pertub_func = KernelShap._kernel_shap_pertub_func,
            ref_value = ref_value,
            map_to_interpret_space = map_to_interpret_space,
            nb_samples = nb_samples,
            )

    # No need to redifine the explain method (herited from Lime)

    @staticmethod
    @tf.function
    def _kernel_shap_pertub_func(nb_features: Union[int, tf.Tensor],
                                 nb_samples: int) -> tf.Tensor:
        """
        The perturbed instances are sampled that way:
         - We choose a number of selected features k, considering the distribution
                p(k) = (nb_features - 1) / (k * (nb_features - k))
            where nb_features is the total number of features in the interpretable space
         - Then we randomly select a binary vector with k ones, all the possible sample
           are equally likely. It is done by generating a random vector with values drawn
           from a normal distribution and keeping the top k elements which then will be 1
           and other values are 0.
         Since there are nb_features choose k vectors with k ones, this weighted sampling
         is equivalent to applying the Shapley kernel for the sample weight, defined as:
            k(nb_features, k) = (nb_features - 1)/(k*(nb_features - k)*(nb_features choose k))
        This trick is the one used in the Captum library: https://github.com/pytorch/captum
        """

        nb_features = tf.squeeze(nb_features)
        probs_nb_selected_feature = KernelShap._get_probs_nb_selected_feature(
            tf.cast(nb_features, dtype=tf.int32)
        )
        nb_selected_features = tf.random.categorical(tf.math.log([probs_nb_selected_feature]),
                                                     nb_samples,
                                                     dtype=tf.int32)
        nb_selected_features = tf.reshape(nb_selected_features, [nb_samples])
        nb_selected_features = tf.one_hot(nb_selected_features, nb_features, dtype=tf.int32)

        rand_vals = tf.random.normal([nb_samples, nb_features])
        idx_sorted_values = tf.argsort(rand_vals, axis=1, direction='DESCENDING')

        threshold_idx = idx_sorted_values * nb_selected_features
        threshold_idx = tf.reduce_sum(threshold_idx, axis=1)

        threshold = rand_vals * tf.one_hot(threshold_idx, nb_features)
        threshold = tf.reduce_sum(threshold, axis=1)
        threshold = tf.expand_dims(threshold, axis=1)
        threshold = tf.repeat(threshold, repeats=nb_features, axis=1)

        interpret_samples = tf.greater(rand_vals, threshold)
        interpret_samples = tf.cast(interpret_samples, dtype=tf.int32)

        return interpret_samples

    @staticmethod
    @tf.function
    def _get_probs_nb_selected_feature(num_features: Union[int, tf.Tensor]) -> tf.Tensor:
        """
        Compute the distribution:
            p(k) = (nb_features - 1) / (k * (nb_features - k))
        """
        list_features_indexes = tf.range(1, num_features)
        denom = tf.multiply(list_features_indexes, (num_features - list_features_indexes))
        num = num_features - 1
        probs = tf.divide(num, denom)
        probs = tf.concat([[0.0], probs], 0)
        return tf.cast(probs, dtype=tf.float32)

    @staticmethod
    def _kernel_shap_similarity_kernel(
        original_input,
        interpret_samples,
        perturbed_samples
    ) -> tf.Tensor:
    # pylint: disable=unused-argument
        """
        This method compute the similarity between interpretable perturbed samples and
        the original input (i.e a tf.ones(num_features)). The trick used for computation
        reason is to instead of using the original similarity kernel to pick random perturbed
        instances of interpretable sample in order to follow a certain probability rule, we
        let the pertub function create the pertub interpretable sample directly following
        this probability. Therefore, from the pertub function we can pick them with
        equal probability. See the `_kernel_shap_pertub_func` for more details.
        """

        similarities = tf.ones(len(interpret_samples), dtype=tf.float32)

        return similarities



import numpy as np
from sklearn.linear_model import LinearRegression
import tensorflow as tf
from tensorflow.keras.models import Sequential, Model
from tensorflow.keras.layers import Dense, Conv1D, Conv2D, Activation, GlobalAveragePooling1D, Dropout, Flatten, MaxPooling2D, Input
from tensorflow.keras.utils import to_categorical

def generate_data(x_shape=(32, 32, 3), num_labels=10, samples=100):
    x = np.random.rand(samples, *x_shape).astype(np.float32)
    y = to_categorical(np.random.randint(0, num_labels, samples), num_labels)

    return x, y

def generate_model(input_shape=(32, 32, 3), output_shape=10):
    model = Sequential()
    model.add(Input(shape=input_shape))
    model.add(Conv2D(4, kernel_size=(2, 2),
                     activation='relu'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Dropout(0.25))
    model.add(Flatten())
    model.add(Dense(output_shape))
    model.add(Activation('softmax'))
    model.compile(loss='categorical_crossentropy', optimizer='sgd')

    return model

def generate_timeseries_model(input_shape=(20, 10), output_shape=10):
    model = Sequential()
    model.add(Input(shape=input_shape))
    model.add(Conv1D(4, kernel_size=3, activation='relu'))
    model.add(GlobalAveragePooling1D())
    model.add(Dense(output_shape))
    model.add(Activation('softmax'))
    model.compile(loss='categorical_crossentropy', optimizer='sgd',
                  metrics=['accuracy'])

    return model

def generate_regression_model(features_shape, output_shape=1):
    model = Sequential()
    model.add(Input(shape=features_shape))
    model.add(Flatten())
    model.add(Dense(4, activation='relu'))
    model.add(Dense(4, activation='relu'))
    model.add(Dense(output_shape))
    model.compile(loss='mean_absolute_error', optimizer='sgd',
                  metrics=['accuracy'])

    return model

def almost_equal(arr1, arr2, epsilon=1e-6):
    """Ensure two array are almost equal at an epsilon"""
    return np.sum(np.abs(arr1 - arr2)) < epsilon


def generate_dataset(nb_samples: int = 100):
    np.random.seed(0)

    # dataset parameters
    features_coef = np.array([1, -2, -3, 4, 0, 0, 0, 0])
    nb_features = len(features_coef)

    # create dataset
    dataset = np.random.normal(4, 1, (nb_samples, nb_features))
    noise = np.random.normal(0, 1, nb_samples)
    target = dataset.dot(features_coef) + noise

    return dataset, target, features_coef


def generate_linear_model(
        features_coef,
        library: str="keras"
):
    if library == "sklearn":
        sk_linear = LinearRegression()

        # fit is necessary for the model to use predict
        data = np.ones((len(features_coef),)).reshape(1, -1)
        target = np.array([0])
        sk_linear.fit(data, target)
        # but it does not impact the coefficients
        sk_linear.coef_ = features_coef

        # Create the wrapper class
        class Wrapper:
            # The init method is necessary for every class
            def __init__(self, model):
                self.model = model

            # The call method calls the predict method
            def __call__(self, inputs):
                pred = self.model.predict(inputs)
                return pred

        sk_model = Wrapper(sk_linear)
        return sk_model

    elif library == "keras":

        inputs = Input(shape=(len(features_coef),))

        # create one dense layer with the weights
        weights = np.expand_dims(np.array(features_coef), axis=1)
        bias = np.array([0])
        output = tf.keras.layers.Dense(
            1, weights=[weights, bias]
        )(inputs)

        # create and compile model
        tf_model = Model(inputs, output)
        tf_model.compile(optimizer="adam", loss="mae")

        return tf_model




"""
Optimisation functions
"""

import tensorflow as tf

from ..commons.model_override import override_relu_gradient, open_relu_policy
from ..types import Optional, Union, List, Callable, Tuple
from .preconditioning import fft_image, get_fft_scale, fft_to_rgb, to_valid_rgb
from .transformations import generate_standard_transformations
from .objectives import Objective


def optimize(objective: Objective,
             optimizer: Optional[tf.keras.optimizers.Optimizer] = None,
             nb_steps: int = 256,
             use_fft: bool = True,
             fft_decay: float = 0.85,
             std: float = 0.01,
             regularizers: Optional[List[Callable]] = None,
             image_normalizer: str = 'sigmoid',
             values_range: Tuple[float, float] = (0, 1),
             transformations: Optional[Union[List[Callable], str]] = 'standard',
             warmup_steps: int = False,
             custom_shape: Optional[Tuple] = (512, 512),
             save_every: Optional[int] = None) -> Tuple[List[tf.Tensor], List[str]]:
             # pylint: disable=R0913,E1130
    """
    Optimise a given objective using gradient ascent.

    Parameters
    ----------
    objective
        Objective object.
    optimizer
        Optimizer used for gradient ascent.
    nb_steps
        Number of iterations.
    use_fft
        If true, use fourier preconditioning.
    fft_decay
        Control the allowed energy of the high frequency, a high value
        suppresses high frequencies.
    std
        Standard deviation used for the image initialization (or buffer for fft).
    regularizers
        List of regularizers that are applied on the image and added to the loss.
    image_normalizer
        Transformation applied to the image after each iterations to ensure the
        pixels are in [0,1].
    values_range
        Range of values of the inputs that will be provided to the model, e.g (0, 1) or (-1, 1).
    transformations
        Transformations applied to the image during optimisation, default to robust standard
        transformations.
    warmup_steps
        If true, clone the model by replacing the Relu's with Leaky Relu's to
        find a pre-optimised image, allowing the visualization process to get
        started (as the relu could block the flow of gradient).
    custom_shape
        If specified, optimizes images of the given size. Often use with
        jittering & scale to optimize bigger images crop by crop.
    save_every
        Define the steps to which we save the optimized images. In any case, the
        last images will be returned.

    Returns
    -------
    images_optimized
        Optimized images for each objectives.
    objective_names
        Name of each objectives.
    """
    values_range = (min(values_range), max(values_range))

    model, objective_function, objective_names, input_shape = objective.compile()

    if optimizer is None:
        optimizer = tf.keras.optimizers.Adam(0.05)

    img_shape = input_shape
    if custom_shape:
        img_shape = (img_shape[0], *custom_shape, img_shape[-1])

    if transformations == 'standard':
        transformations = generate_standard_transformations(img_shape[1])

    if use_fft:
        inputs = tf.Variable(fft_image(img_shape, std), trainable=True)
        fft_scale = get_fft_scale(img_shape[1], img_shape[2], decay_power=fft_decay)
        image_param = lambda inputs: to_valid_rgb(fft_to_rgb(img_shape, inputs, fft_scale),
                                                  image_normalizer, values_range)
    else:
        inputs = tf.Variable(tf.random.normal(img_shape, std, dtype=tf.float32))
        image_param = lambda inputs: to_valid_rgb(inputs, image_normalizer, values_range)

    optimisation_step = _get_optimisation_step(objective_function,
                                               len(model.outputs),
                                               image_param,
                                               input_shape,
                                               transformations,
                                               regularizers)

    if warmup_steps:
        model_warmup = override_relu_gradient(model, open_relu_policy)
        for _ in range(warmup_steps):
            grads = optimisation_step(model_warmup, inputs)
            optimizer.apply_gradients([(-grads, inputs)])

    images_optimized = []
    for step_i in range(nb_steps):
        grads = optimisation_step(model, inputs)
        optimizer.apply_gradients([(-grads, inputs)])

        last_iteration = step_i == nb_steps - 1
        should_save = save_every and (step_i + 1) % save_every == 0
        if should_save or last_iteration:
            imgs = image_param(inputs)
            images_optimized.append(imgs)

    return images_optimized, objective_names


def _get_optimisation_step(
        objective_function: Callable,
        nb_outputs: int,
        image_param: Callable,
        input_shape: Tuple,
        transformations: Optional[Callable] = None,
        regularizers: Optional[List[Callable]] = None) -> Callable:
    """
    Generate a function that optimize the objective function for a single step.

    Parameters
    ----------
    objective_function
        Function that compute the loss for the objectives given the model
        outputs.
    nb_outputs
        Number of outputs of the model.
    image_param
        Function that map image to a valid rgb.
    input_shape
        Shape of the inputs to optimize.
    transformations
        Transformations applied to the image during optimisation.
    regularizers
        List of regularizers that are applied on the image and added to the loss.

    Returns
    -------
    step_function
        Function (model, inputs) to call to optimize the input for one step.
    """

    @tf.function
    def step(model, inputs):

        with tf.GradientTape() as tape:
            tape.watch(inputs)

            imgs = image_param(inputs)
            if transformations:
                imgs = transformations(imgs)
            imgs = tf.image.resize(imgs, (input_shape[1], input_shape[2]))

            model_outputs = model(imgs)

            if nb_outputs == 1:
                model_outputs = tf.expand_dims(model_outputs, 0)

            loss = objective_function(model_outputs)

            if regularizers:
                for reg_function in regularizers:
                    loss -= reg_function(imgs)

        grads = tape.gradient(loss, inputs)

        return grads

    return step



"""
Module related to Object detector method
"""

from typing import Iterable, Tuple, Union, Optional
import abc

import tensorflow as tf
import numpy as np

from xplique.attributions.base import BlackBoxExplainer
from xplique.commons import operator_batching


class IouCalculator:
    """
    Used to compute the Intersection Over Union (IOU).
    """
    @abc.abstractmethod
    def intersect(self, objects_a: tf.Tensor, objects_b: tf.Tensor) -> tf.Tensor:
        """
        Compute the intersection between two batched objects (e.g boxes, segmentation masks...)

        Parameters
        ----------
        objects_a
            First batch of objects to compare.
        objects_b
            Second batch of objects to compare.

        Returns
        -------
        score
            Real value between [0,1] corresponding to the intersection of the 2 objects.
        """
        raise NotImplementedError()


class SegmentationIouCalculator(IouCalculator):
    """
    Compute segmentation masks IOU.
    """
    def intersect(self, masks_a: tf.Tensor, masks_b: tf.Tensor) -> tf.Tensor:
        """
        Compute the intersection between two batched segmentation masks.
        Each segmentation is a boolean mask on the whole image

        Parameters
        ----------
        masks_a
            First batch of segmentation masks.
        masks_b
            Second batch of segmentation masks.

        Returns
        -------
        iou_score
            The IOU score between the first and second batch of masks.
        """
        # pylint: disable=W0221,W0237

        axis = np.arange(1, tf.rank(masks_a))

        inter_area = tf.reduce_sum(tf.cast(tf.logical_and(masks_a, masks_b),
                                    dtype=tf.float32), axis=axis)
        union_area = tf.reduce_sum(tf.cast(tf.logical_or(masks_a, masks_b),
                                    dtype=tf.float32), axis=axis)

        iou_score = inter_area / tf.maximum(union_area, 1.0)

        return iou_score



class BoxIouCalculator(IouCalculator):
    """
    Used to compute the Bounding Box IOU.
    """
    EPSILON = tf.constant(1e-4)

    def intersect(self, boxes_a: tf.Tensor, boxes_b: tf.Tensor) -> tf.Tensor:
        """
        Compute the intersection between two batched bounding boxes.
        Each bounding box is defined by (x1, y1, x2, y2) respectively (left, bottom, right, top).

        Parameters
        ----------
        boxes_a
            First batch of bounding boxes.
        boxes_b
            Second batch of bounding boxes.

        Returns
        -------
        iou_score
           The IOU score between the first and second batch of bounding boxes.
        """
        # pylint: disable=W0221,W0237

        # determine the intersection rectangle
        left   = tf.maximum(boxes_a[..., 0], boxes_b[..., 0])
        bottom = tf.maximum(boxes_a[..., 1], boxes_b[..., 1])
        right  = tf.minimum(boxes_a[..., 2], boxes_b[..., 2])
        top    = tf.minimum(boxes_a[..., 3], boxes_b[..., 3])

        intersection_area = tf.math.maximum(right - left, 0) * tf.math.maximum(top - bottom, 0)

        # determine the areas of the prediction and ground-truth rectangles
        a_area = (boxes_a[..., 2] - boxes_a[..., 0]) * (boxes_a[..., 3] - boxes_a[..., 1])
        b_area = (boxes_b[..., 2] - boxes_b[..., 0]) * (boxes_b[..., 3] - boxes_b[..., 1])

        union_area = a_area + b_area - intersection_area

        iou_score = intersection_area / (union_area + BoxIouCalculator.EPSILON)

        return iou_score


class IObjectFormater:
    """
    Generic class to format the model prediction
    """
    def format_objects(self, predictions) -> Iterable[Tuple[tf.Tensor, tf.Tensor, tf.Tensor]]:
        """
        Format the model prediction of a given image to have the prediction of the following format:
        objects, proba_detection, one_hots_classifications

        Parameters
        ----------
        predictions
            prediction of the model of a given image

        Returns
        -------
        object
            bounding box or mask component of the prediction
        proba
            existence probability component of the prediction
        classification
            classification component of the prediction
        """
        raise NotImplementedError()


class ImageObjectDetectorScoreCalculator:
    """
    Class to compute batch score
    """
    def __init__(self, object_formater: IObjectFormater, iou_calculator: IouCalculator):
        self.object_formater = object_formater
        self.iou_calculator = iou_calculator

        self.batch_score = operator_batching(self.score)

    def score(self, model, inp, object_ref) -> tf.Tensor:
        """
        Compute the matching score between prediction and a given object

        Parameters
        ----------
        model
            the model used for the object detection
        inp
            the batched image
        object_ref
            the object target to compare with the prediction of the model

        Returns
        -------
        score
            for each image, the matching score between the object of reference and
            the prediction of the model
        """
        objects = model(inp)
        score_values = []
        for obj, obj_ref in zip(objects, object_ref):
            if obj is None or obj.shape[0] == 0:
                score_values.append(tf.constant(0.0, dtype=inp.dtype))
            else:
                current_boxes, proba_detection, classification = \
                            self.object_formater.format_objects(obj)

                if len(tf.shape(obj_ref)) == 1:
                    obj_ref = tf.expand_dims(obj_ref, axis=0)

                obj_ref = self.object_formater.format_objects(obj_ref)

                scores = []
                size = tf.shape(current_boxes)[0]
                for boxes_ref, proba_ref, class_ref in zip(*obj_ref):
                    boxes_ref = tf.repeat(tf.expand_dims(boxes_ref, axis=0), repeats=size, axis=0)
                    proba_ref = tf.repeat(tf.expand_dims(proba_ref, axis=0), repeats=size, axis=0)
                    class_ref = tf.repeat(tf.expand_dims(class_ref, axis=0), repeats=size, axis=0)

                    iou = self.iou_calculator.intersect(boxes_ref, current_boxes)
                    classification_similarity = tf.reduce_sum(class_ref * classification, axis=1) \
                            / (tf.norm(classification, axis=1) * tf.norm(class_ref, axis=1))

                    current_score = iou * tf.squeeze(proba_detection, axis=1) \
                                        * classification_similarity
                    current_score = tf.reduce_max(current_score)
                    scores.append(current_score)

                score_value = tf.reduce_max(tf.stack(scores))
                score_values.append(score_value)

        score_values = tf.stack(score_values)

        return score_values


class ImageObjectDetectorExplainer(BlackBoxExplainer):
    """
    Used to define method as an object detector one
    """

    def __init__(self, explainer: BlackBoxExplainer, object_detector_formater: IObjectFormater,
                 iou_calculator: IouCalculator):
        """
        Constructor

        Parameters
        ----------
        explainer
            the black box explainer used to explain the object detector model
        object_detector_formater
            the formater of the object detector model used to format the prediction
            of the right format
        iou_calculator
            the iou calculator used to compare two objects.
        """
        super().__init__(explainer.model, explainer.batch_size)
        self.explainer = explainer
        self.score_calculator = ImageObjectDetectorScoreCalculator(object_detector_formater,
                                                                    iou_calculator)
        self.explainer.inference_function = self.score_calculator.score
        self.explainer.batch_inference_function = self.score_calculator.batch_score

    def explain(self, inputs: Union[tf.data.Dataset, tf.Tensor, np.array],
                targets: Optional[Union[tf.Tensor, np.array]] = None) -> tf.Tensor:
        if len(tf.shape(targets)) == 1:
            targets = tf.expand_dims(targets, axis=0)

        return self.explainer.explain(inputs, targets)


class BoundingBoxesExplainer(ImageObjectDetectorExplainer, IObjectFormater):
    """
    For a given black box explainer, this class allows to find explications of an object detector
    model. The object model detector shall return a list (length of the size of the batch)
    containing a tensor of 2 dimensions.
    The first dimension of the tensor is the number of bounding boxes found in the image
    The second dimension is:
    [x1_box, y1_box, x2_box, y2_box, probability_detection, ones_hot_classif_result]

    This work is a generalisation of the following article at any kind of black box explainer and
    also can be used for other kind of object detector (like segmentation)

    Ref. Petsiuk & al., Black-box Explanation of Object Detectors via Saliency Maps (2021).
    https://arxiv.org/pdf/2006.03204.pdf
    """

    def __init__(self, explainer: BlackBoxExplainer):
        super().__init__(explainer, self, BoxIouCalculator())

    def format_objects(self, predictions) -> Iterable[Tuple[tf.Tensor, tf.Tensor, tf.Tensor]]:
        boxes, proba_detection, one_hots_classifications = tf.split(predictions,
                                                    [4, 1, tf.shape(predictions[0])[0] - 5], 1)
        return boxes, proba_detection, one_hots_classifications



import tensorflow as tf
import numpy as np

from tensorflow.keras.optimizers import SGD

from xplique.features_visualizations import (Objective, optimize, l1_reg, l2_reg,
                                             total_variation_reg, compose_transformations, pad,
                                             random_jitter, random_blur, random_flip, random_scale)



def dummy_model():
    model = tf.keras.Sequential([
        tf.keras.layers.Input((16, 16, 3)),
        tf.keras.layers.Conv2D(3, (3, 3), name="early"),
        tf.keras.layers.MaxPool2D((2, 2)),
        tf.keras.layers.Conv2D(3, (3, 3), name="features"),
        tf.keras.layers.MaxPool2D((2, 2)),
        tf.keras.layers.Flatten(name="pre-logits"),
        tf.keras.layers.Dense(5, name="logits")
    ])
    model.compile()
    return model

def _assert_objective(obj):
    """ Assert the objective return appropriate results under multiple setting """
    regularizers = [l1_reg(1.0), l2_reg(1.0), total_variation_reg(1.0)]
    transformations = compose_transformations([
                    pad(2, 0.0),
                    random_jitter(6),
                    random_scale(scale_range=[0.95, 1.05]),
                    random_blur(),
                    random_flip()
    ])
    nb_steps = 4

    for use_fourier in [True, False]:
        for custom_shape in [None, (20, 20)]:
            for warmup_steps in [False, 2]:
                for save_every in [None, 2]:

                    res, _ = optimize(obj,
                                      SGD(),
                                      nb_steps=nb_steps,
                                      use_fft=use_fourier,
                                      regularizers=regularizers,
                                      warmup_steps=warmup_steps,
                                      custom_shape=custom_shape,
                                      transformations=transformations,
                                      save_every=save_every)

                    if save_every is not None:
                        assert len(res) == nb_steps // 2
                    if custom_shape is not None:
                        assert res[0].shape[1:] == (*custom_shape, 3)
                    else:
                        assert res[0].shape[1:] == (16, 16, 3)


def test_layer():
    """ Ensure we can optimize on a layer """
    model = dummy_model()

    obj = Objective.layer(model, -1)
    _assert_objective(obj)


def test_direction():
    """ Ensure we can optimize on a direction """
    model = dummy_model()

    direction_vec = np.random.random(model.get_layer("early").output.shape[1:])
    obj = Objective.direction(model, "early", direction_vec)
    _assert_objective(obj)


def test_channel():
    """ Ensure we can optimize on a channel """
    model = dummy_model()
    nb_channels = 2 # number of channel to optim (1...n)

    obj = Objective.channel(model, "early", list(range(nb_channels)))
    _assert_objective(obj)


def test_neurons():
    """ Ensure we can optimize on neurons """
    model = dummy_model()

    obj = Objective.neuron(model, "early", list(range(3)))
    _assert_objective(obj)


def test_combinations():
    """ Ensure we can optimize a combinations of objectives """
    model = dummy_model()

    layer_obj = Objective.layer(model, -1)

    direction_vec = np.random.random(model.get_layer("pre-logits").output.shape[1:])
    direction_obj = Objective.direction(model, "pre-logits", direction_vec)

    nb_channels = 2
    channels_obj = Objective.channel(model, "early", list(range(nb_channels)))

    nb_neurons = 3
    neurons_obj = Objective.neuron(model, "features", list(range(nb_neurons)))

    combined_obj = layer_obj + direction_obj + channels_obj + neurons_obj
    _assert_objective(combined_obj)



"""
Tests that BlackBox attribution model can indeed handle
several model type with a native User eXperience
"""

import numpy as np
import tensorflow as tf

from sklearn.svm import SVC
from sklearn.ensemble import RandomForestClassifier

from xplique.attributions import (Occlusion, Rise, Lime, KernelShap, SobolAttributionMethod)
from xplique.commons.operators import predictions_operator, batch_predictions,\
    batch_predictions_one_hot_callable
from xplique.commons.callable_operations import predictions_one_hot_callable

from ..utils import generate_data, generate_model, generate_regression_model

def _default_methods_tabular(model):
    return [
        Occlusion(model, patch_size = 1, patch_stride = 1),
        Lime(model),
        KernelShap(model)
    ]

def _default_methods_images(model):
    return [
        Rise(model),
        Occlusion(model),
        Lime(model, map_to_interpret_space=_map_to_interpret_space),
        KernelShap(model, map_to_interpret_space=_map_to_interpret_space),
        SobolAttributionMethod(model, grid_size=2, nb_design=2)
    ]

def _map_to_interpret_space(inp):
    width = inp.shape[0]
    height = inp.shape[1]

    mapping = tf.range(width*height)
    mapping = tf.reshape(mapping, (width, height))
    mapping = tf.cast(mapping, tf.int32)

    return mapping

class DenseModule(tf.Module):
    """
    A generic tf.Module Dense layer
    """
    def __init__(self, input_size, nb_labels):
        super(DenseModule, self).__init__()
        self.weights = tf.Variable(
            tf.random.normal([input_size, nb_labels]), name='w')
        self.bias = tf.Variable(tf.zeros([nb_labels]), name='b')
    def __call__(self, x):
        output = tf.matmul(x, self.weights) + self.bias
        return tf.nn.relu(output)

class ImagesModule(tf.Module):
    """
    A generic tf.Module with flatten and two dense layers
    """
    def __init__(self, input_shape, nb_labels):
        super(ImagesModule, self).__init__()
        self.input_size = input_shape[0]*input_shape[1]*input_shape[2]
        self.dense1 = DenseModule(self.input_size, 64)
        self.dense2 = DenseModule(64, nb_labels)
    def __call__(self, x):
        x = tf.reshape(x, [x.shape[0],-1])
        x = self.dense1(x)
        output = self.dense2(x)
        return output

def _default_tf_model_tabular(input_size, nb_labels):
    tf_keras_model = generate_regression_model((input_size,), nb_labels)
    tf_keras_layer = tf.keras.layers.Dense(nb_labels)
    tf_keras_layer(tf.random.normal([32, input_size]))
    tf_module = DenseModule(input_size, nb_labels)

    return [
        tf_keras_model,
        tf_keras_layer,
        tf_module
    ]

def _default_tf_model_images(input_shape, nb_labels):

    tf_keras_model = generate_model(input_shape, nb_labels)

    # create a mode through tf functionnal api
    functional_inputs = tf.keras.Input(shape = input_shape)
    functional_flatten = tf.keras.layers.Flatten()
    functional_x = functional_flatten(functional_inputs)
    functional_x = tf.keras.layers.Dense(64, activation="relu")(functional_x)
    functional_outputs = tf.keras.layers.Dense(nb_labels)(functional_x)
    tf_functional_model = tf.keras.Model(inputs=functional_inputs, outputs=functional_outputs)

    tf_module = ImagesModule(input_shape, nb_labels)

    return [
        tf_keras_model,
        tf_functional_model,
        tf_module
    ]

def _default_callable_tabular(input_size, nb_labels):
    sk_svc = SVC(probability=True)
    sk_rf = RandomForestClassifier()
    sk_api = RandomSklearn(input_size, nb_labels)
    np_callable = RandomNpCallable(input_size, nb_labels)

    return [
        sk_svc,
        sk_rf,
        sk_api,
        np_callable
    ]

def _default_callable_images(input_shape, nb_labels):
    sk_api = RandomSklearn(input_shape, nb_labels)
    np_callable = RandomNpCallable(input_shape, nb_labels)

    return [
        sk_api,
        np_callable
    ]

class RandomSklearn():
    """
    A class with a fit and predict_proba attributes such as the one
    in sklearn api, which returns random probabilities
    """
    def __init__(self, input_shape, nb_labels):
        self.input_shape = input_shape
        self.nb_labels = nb_labels
    def fit(self, inputs, targets):
        pass
    def predict_proba(self, inputs):
        return np.random.random((inputs.shape[0], self.nb_labels))

class RandomNpCallable():
    """
    A class with a call defines on numpy array
    """
    def __init__(self, input_shape, nb_labels):
        self.input_shape = input_shape
        self.nb_labels = nb_labels
    def fit(self, inputs, targets):
        pass
    def __call__(self, inputs):
        return np.random.random((inputs.shape[0], self.nb_labels))

def test_tf_models_tabular():
    """
    Test if on tabular data, BlackBox methods work as expected with tf family of callable,
    at least the most commons.
    """
    input_size = 32
    nb_labels = 10
    samples = 10

    inputs, targets = generate_data((input_size,), nb_labels, samples)
    models = _default_tf_model_tabular(input_size, nb_labels)

    for model in models:
        explainers = _default_methods_tabular(model)

        for explainer in explainers:
            assert explainer.inference_function is predictions_operator
            assert explainer.batch_inference_function is batch_predictions

            explanations = explainer(inputs, targets)

            assert explanations.shape == (samples, input_size)

def test_tf_models_images():
    """
    Test if on images data, BlackBox methods work as expected with tf family of callable,
    at least the most commons.
    """
    input_shapes = [(28, 28, 1), (32, 32, 3)]
    nb_labels = 10
    samples = 10

    for input_shape in input_shapes:
        inputs, targets = generate_data(input_shape, nb_labels, samples)
        models = _default_tf_model_images(input_shape, nb_labels)
        for model in models:

            explainers = _default_methods_images(model)

            for explainer in explainers:
                assert explainer.inference_function is predictions_operator
                assert explainer.batch_inference_function is batch_predictions

                explanations = explainer(inputs, targets)

                assert explanations.shape == (samples, *input_shape[:2])

def test_callable_models_tabular():
    """
    Test if on tabular data we can handle sklearn models, sklearn like models (with
    predict_proba attributes) and callable which requires a np.ndarray input
    """
    input_size = 32
    nb_labels = 10
    samples = 10

    inputs, targets = generate_data((input_size,), nb_labels, samples)
    models = _default_callable_tabular(input_size, nb_labels)

    for model in models:
        no_one_hot_targets = np.arange(0,10)

        model.fit(inputs, no_one_hot_targets)
        explainers = _default_methods_tabular(model)

        for explainer in explainers:
            assert explainer.inference_function is predictions_one_hot_callable
            assert explainer.batch_inference_function is batch_predictions_one_hot_callable

            explanations = explainer(inputs, targets)

            assert explanations.shape == (samples, input_size)

def test_callable_models_images():
    """
    Test if on images data we can handle sklearn like models (with predict_proba attributes)
    and callable which requires a np.ndarray input
    """
    input_shapes = [(28, 28, 1), (32, 32, 3)]
    nb_labels = 10
    samples = 10

    for input_shape in input_shapes:
        inputs, targets = generate_data(input_shape, nb_labels, samples)
        models = _default_callable_images(input_shape, nb_labels)
        for model in models:

            explainers = _default_methods_images(model)

            for explainer in explainers:
                assert explainer.inference_function is predictions_one_hot_callable
                assert explainer.batch_inference_function is batch_predictions_one_hot_callable

                explanations = explainer(inputs, targets)

                assert explanations.shape == (samples, *input_shape[:2])



"""
Module related to LIME method
"""

import warnings

import tensorflow as tf
import numpy as np
from tensorflow.keras.losses import cosine_similarity #pylint:  disable=E0611
from sklearn import linear_model
from skimage.segmentation import quickshift, felzenszwalb

from .base import BlackBoxExplainer, sanitize_input_output
from ..commons import Tasks
from ..types import Callable, Union, Optional, Any, OperatorSignature

class Lime(BlackBoxExplainer):
    """
    Used to compute the LIME method.

    Ref. Ribeiro & al., "Why Should I Trust You?": Explaining the Predictions of Any Classifier.
    https://arxiv.org/abs/1602.04938

    Notes
    -----

    Note that the quality of an explanation relies strongly on your choice of the interpretable
    model, the similarity kernel and the map function mapping a sample into an interpretable space.
    The similarity kernel will define how close the perturbed samples are from the original sample
    you want to explain.
    For instance, if you have large images (e.g 299x299x3) a similarity kernel with the argument
    kernel width of 1 will compute similarities really close to 0. Consequently the interpretable
    model will not train. In order to make it work you have to use (for example) a larger kernel
    width.
    Moreover, depending on the similarities vector some interpretable model will fit
    better than other (e.g Ridge on large colored image might perform better than Lasso).
    Finally, your map function will defines how many features your linear model has to learn.
    Basically, an identity mapping (map each pixel as a single feature), on large image means there
    is as many features as there are pixels (e.g 299x299x3->89401 features) which can lead to poor
    explanations.

    N.B: This module was built to be deployed on GPU to be fully efficient. Considering the number
    of samples and number of inputs you want to process it might even be necessary.

    Parameters
    ----------
    model
        The model from which we want to obtain explanations
    batch_size
        Number of perturbed samples to process at once, mandatory when nb_samples is huge.
        Notice, it is different compare to WhiteBox explainers which batch the inputs.
        Here inputs are process one by one.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].
    interpretable_model
        Model object to train interpretable model.
        See the documentation for more information.
    similarity_kernel
        Function which considering an input, perturbed instances of these input and
        the interpretable version of those perturbed samples compute the similarities between
        the input and the perturbed samples.
        See the documentation for more information.
    pertub_func
        Function which generate perturbed interpretable samples in the interpretation space from
        the number of interpretable features (e.g nb of super pixel) and the number of perturbed
        samples you want per original input.
        See the documentation for more information.
    ref_value
        It defines reference value which replaces each feature when the corresponding
        interpretable feature is set to 0.
        It should be provided as: a ndarray of shape (1) if there is no channels in your input
        and (C,) otherwise

        The default ref value is set to (0.5,0.5,0.5) for inputs with 3 channels (corresponding
        to a grey pixel when inputs are normalized by 255) and to 0 otherwise.
    map_to_interpret_space
        Function which group features of an input corresponding to the same interpretable
        feature (e.g super-pixel).
        It allows to transpose from (resp. to) the original input space to (resp. from)
        the interpretable space.
        See the documentation for more information.
    nb_samples
        The number of perturbed samples you want to generate for each input sample.
        Default to 150.
    prob
        The probability argument for the default pertub function.
    distance_mode
        The distance mode used in the default similarity kernel, you can choose either
        "euclidean" or "cosine" (will compute cosine similarity).
        Default value set to "euclidean".
    kernel_width
        Width of your kernel. It is important to make it evolving depending on your inputs size
        otherwise you will get all similarity close to 0 leading to poor performance or NaN
        values.
        Default to 45 (i.e adapted for RGB images).
    """
    def __init__(
        self,
        model: Callable,
        batch_size: Optional[int] = None,
        operator: Optional[Union[Tasks, str, OperatorSignature]] = None,
        interpretable_model: Any = linear_model.Ridge(alpha=2),
        similarity_kernel: Optional[Callable[[tf.Tensor, tf.Tensor, tf.Tensor], tf.Tensor]] = None,
        pertub_func: Optional[Callable[[Union[int, tf.Tensor],int], tf.Tensor]] = None,
        map_to_interpret_space: Optional[Callable[[tf.Tensor], tf.Tensor]] = None,
        ref_value: Optional[np.ndarray] = None,
        nb_samples: int = 150,
        distance_mode: str = "euclidean",
        kernel_width: float = 45.0,
        prob: float = 0.5
        ): # pylint: disable=R0913

        if not all(hasattr(interpretable_model, attr) for attr in ['fit', 'predict']):
            raise ValueError(
                "The interpretable model is invalid. It should have a fit and a predict"
                " method. It should also have a coef_ attribute (the interpretable "
                "explanation) once fit is called."
                )

        if similarity_kernel is None:
            similarity_kernel = Lime._get_exp_kernel_func(distance_mode, kernel_width)

        if pertub_func is None:
            pertub_func = Lime._get_default_pertub_function(prob)

        if (nb_samples>=500) and (batch_size is None):
            warnings.warn(
                "You set a number of perturbed samples per input >= 500 and "
                "batch_size is set to None"
                "This mean that you will ask your model to handle more than 500"
                " perturbed samples per input at once."
                "This can lead to OOM issue. To avoid it you can set the"
                " batch_size argument."
            )

        super().__init__(model, batch_size, operator)

        self.map_to_interpret_space = map_to_interpret_space
        self.interpretable_model = interpretable_model
        self.similarity_kernel = similarity_kernel
        self.pertub_func = pertub_func
        self.ref_value = ref_value
        self.nb_samples = nb_samples

    @sanitize_input_output
    def explain(self,
                inputs: Union[tf.data.Dataset, tf.Tensor, np.ndarray],
                targets: Optional[Union[tf.Tensor, np.ndarray]] = None) -> tf.Tensor:
        """
        This method attributes the output of the model with given targets
        to the inputs of the model using the approach described above,
        training an interpretable model and returning a representation of the
        interpretable model.

        Parameters
        ----------
        inputs
            Dataset, Tensor or Array. Input samples to be explained.
            If Dataset, targets should not be provided (included in Dataset).
            Expected shape among (N, W), (N, T, W), (N, H, W, C).
            More information in the documentation.
        targets
            Tensor or Array. One-hot encoding of the model's output from which an explanation
            is desired. One encoding per input and only one output at a time. Therefore,
            the expected shape is (N, output_size).
            More information in the documentation.

        Returns
        -------
        explanations
            Interpretable coefficients, same shape as the inputs, except for the channels.
            Coefficients of the interpretable model. Those coefficients having the size of the
            interpretable space will be given the same value to coefficient which were grouped
            together (e.g belonging to the same super-pixel).
        """

        # check if inputs are tabular or has shape (N, H, W, C)
        is_tabular = len(inputs.shape) == 2
        has_channels = len(inputs.shape )== 4

        if has_channels:
            # default quickshift segmentation for image
            if self.map_to_interpret_space is None:
                self.map_to_interpret_space = Lime._default_image_map_to_interpret_space
            # if inputs have channels ensure
            if self.ref_value is None:
                if inputs.shape[-1] == 3:
                    # grey pixel
                    ref_value = tf.ones(inputs.shape[-1])*0.5
                else:
                    ref_value = tf.zeros(inputs.shape[-1])
            else:
                assert(
                    self.ref_value.shape[0] == inputs.shape[-1]
                ),"The dimension of ref_values must match inputs (C, )"
                ref_value = tf.cast(self.ref_value, tf.float32)
        else:
            if self.map_to_interpret_space is None:
                if is_tabular:
                    self.map_to_interpret_space = Lime._default_tab_map_to_interpret_space
                else:
                    self.map_to_interpret_space = Lime._default_2dimage_map_to_interpret_space

            if self.ref_value is None:
                ref_value = tf.zeros(1)
            else:
                ref_value = tf.cast(self.ref_value, tf.float32)

        batch_size = self.batch_size or self.nb_samples

        return Lime._compute(self.model,
                            batch_size,
                            inputs,
                            targets,
                            self.inference_function,
                            self.interpretable_model,
                            self.similarity_kernel,
                            self.pertub_func,
                            ref_value,
                            self.map_to_interpret_space,
                            self.nb_samples,
                            )

    @staticmethod
    def _compute(model: Callable,
                batch_size: int,
                inputs: tf.Tensor,
                targets: tf.Tensor,
                inference_function: Callable,
                interpretable_model: Callable,
                similarity_kernel: Callable[[tf.Tensor, tf.Tensor, tf.Tensor], tf.Tensor],
                pertub_func: Callable[[Union[int, tf.Tensor],int], tf.Tensor],
                ref_value: tf.Tensor,
                map_to_interpret_space: Callable[[tf.Tensor], tf.Tensor],
                nb_samples: int,
                ) -> tf.Tensor:
                # pylint: disable=R0913
        """
        This method attributes the output of the model with given targets
        to the inputs of the model using the approach described above,
        training an interpretable model and returning a representation of the
        interpretable model.

        Parameters
        ----------
        model
            The model from which we want to obtain explanations
        inputs
            Dataset, Tensor or Array. Input samples to be explained.
            If Dataset, targets should not be provided (included in Dataset).
            Expected shape among (N, W), (N, T, W), (N, H, W, C).
            More information in the documentation.
        targets
            Tensor or Array. One-hot encoding of the model's output from which an explanation
            is desired. One encoding per input and only one output at a time.
            More information in the documentation.
        inference_function
            Function that allows to get the probability output of the model
        interpretable_model
            Model object to train interpretable model.
        similarity_kernel
            Function which considering an input, perturbed instances of thoses samples and the
            interpretable version of those perturbed samples compute the similarities.
        pertub_function
            Function which generate perturbed interpretable samples in the interpretation space.
        ref_values
            It defines reference value which replaces each feature when the corresponding
            interpretable feature is set to 0.
        map_to_interpret_space
            Function which group features of an input corresponding to the same interpretable
            feature (e.g super-pixel).
        nb_samples
            The number of perturbed samples you want to generate for each input sample.

        Returns
        -------
        explanations
            A Tensor
            Coefficients of the interpretable model. Those coefficients having the size of the
            interpretable space will be given the same value to coefficient which were grouped
            together (e.g belonging to the same super-pixel).
        """
        explanations = []

        for inp, target in tf.data.Dataset.from_tensor_slices(
            (inputs, targets)
        ):
            # get the mapping of the current input
            mapping = map_to_interpret_space(inp)
            # get the number of interpretable feature
            num_features = tf.reduce_max(mapping) + tf.ones(1, dtype=tf.int32)
            if tf.greater(num_features, 10000):
                warnings.warn(
                    "The current input got a number of interpretable features > 10000. "
                    "This can be very slow or lead to OOM issues when fitting the interpretable "
                    "model. You should consider using a map function which select less features."
                )

            # get perturbed interpretable samples of the input
            interpret_samples = pertub_func(num_features, nb_samples)

            # get the perturbed targets value and the similarities value
            perturbed_targets = []
            similarities = []
            for int_samples in tf.data.Dataset.from_tensor_slices(
                interpret_samples
            ).batch(batch_size):

                masks = Lime._get_masks(int_samples, mapping)
                perturbed_samples = Lime._apply_masks(inp, masks, ref_value)

                augmented_target = tf.expand_dims(target, axis=0)
                augmented_target = tf.repeat(augmented_target, len(perturbed_samples), axis=0)

                batch_perturbed_targets = inference_function(model,
                                                            perturbed_samples,
                                                            augmented_target)

                perturbed_targets.append(batch_perturbed_targets)

                batch_similarities = similarity_kernel(inp, int_samples, perturbed_samples)
                similarities.append(batch_similarities)

            perturbed_targets = tf.concat(perturbed_targets, axis=0)
            similarities = tf.concat(similarities, axis=0)

            # train the interpretable model
            explain_model = interpretable_model

            explain_model.fit(
                interpret_samples.numpy(),
                perturbed_targets.numpy(),
                sample_weight=similarities.numpy()
            )

            explanation = explain_model.coef_
            # cast the interpretable explanation
            explanation = tf.cast(explanation, dtype=tf.float32)

            # broadcast explanations to match the original inputs shapes
            # except for channels
            explanation = Lime._broadcast_explanation(explanation, mapping)

            explanations.append(explanation)

        explanations = tf.stack(explanations, axis=0)

        return explanations

    @staticmethod
    @tf.function(
        input_signature=(
            tf.TensorSpec(shape=[None, None], dtype=tf.int32),
            tf.TensorSpec(shape=None, dtype=tf.int32)
        )
    )
    def _get_masks(interpret_samples: tf.Tensor, mapping: tf.Tensor) -> tf.Tensor:
        """
        This method translate the generated samples in the interpretable space of an input into
        masks to apply to the original input to obtain samples in the original input space.

        Parameters
        ----------
        interpret_samples
            Tensor of shape (nb_samples, num_features)
            Intrepretable samples of an input, with:
                nb_samples number of samples
                num_features the dimension of the interpretable space.
        mapping
            Tensor
            The mapping of the original input from which we drawn interpretable samples.
            Its size is equal to width (and height) of the original input

        Returns
        -------
        masks
            The masks corresponding to each interpretable samples
        """
        tf_masks = tf.gather(interpret_samples, indices=mapping, axis=1)
        return tf_masks

    @staticmethod
    @tf.function
    def _apply_masks(original_input: tf.Tensor,
                     sample_masks: tf.Tensor,
                     ref_value: tf.Tensor) -> tf.Tensor:
        """
        This method apply masks obtained from the perturbed interpretable samples to the
        original input (i.e we get perturbed samples in the original space).

        Parameters
        ----------
        original_input
            The input we want to explain
        sample_masks
            The masks we obtained from the perturbed instances in the interpretable space
        ref_value
            The reference value which replaces each feature when the corresponding
            interpretable feature is set to 0

        Returns
        -------
        perturbed_samples
            The perturbed samples corresponding to the masks applied to the original input
        """
        pert_samples = tf.expand_dims(original_input, axis=0)
        pert_samples = tf.repeat(pert_samples, repeats=sample_masks.shape[0], axis=0)

        # if there is channels we need to expand masks dimension
        if len(original_input.shape)==3:

            sample_masks = tf.expand_dims(sample_masks, axis=-1)
            sample_masks = tf.repeat(sample_masks, repeats=original_input.shape[-1], axis=-1)

            pert_samples = pert_samples * tf.cast(sample_masks, tf.float32)
            ref_val = tf.reshape(ref_value, (1,1,1,original_input.shape[-1]))

        else:

            pert_samples = pert_samples * tf.cast(sample_masks, tf.float32)
            ref_val = tf.reshape(ref_value, (1, *(1,) * len(original_input.shape)))

        pert_samples += (tf.ones((sample_masks.shape)) - tf.cast(sample_masks, tf.float32))*ref_val

        return pert_samples

    @staticmethod
    @tf.function(
        input_signature=(
            tf.TensorSpec(shape=None, dtype=tf.float32),
            tf.TensorSpec(shape=None, dtype=tf.int32)
        )
    )
    def _broadcast_explanation(explanation: tf.Tensor, mapping: tf.Tensor) -> tf.Tensor:
        """
        This method allows to broadcast explanations from the interpretable space to the
        corresponding super pixels

        Parameters
        ----------
        explanation
            Explanation value for each super pixel
        mapping
            The mapping of the original input from which we drawn interpretable samples
            (i.e features index).

        Returns
        -------
        broadcast_explanation
            The explanation of the current input considered
        """

        broadcast_explanation = tf.gather(explanation, indices=mapping, axis=0)
        return broadcast_explanation

    @staticmethod
    def _get_default_pertub_function(
        prob: float = 0.5
        ) -> Callable[[Union[int, tf.Tensor],int], tf.Tensor]:
        """
        This method allows you to get a pertub function with the corresponding prob
        argument.
        """

        prob = tf.cast(prob, dtype=tf.float32)
        @tf.function
        def _default_pertub_function(num_features: Union[int, tf.Tensor],
                                     nb_samples: int) -> tf.Tensor:
            """
            This method generate nb_samples tensor belonging to {0,1}^num_features.
            The prob argument is the probability to have a 1.

            Parameters
            ----------
            num_features
                The number of interpretable features (e.g super pixel).
            nb_samples
                The number of perturbed interpretable samples we want
            prob:
                It defines the probability to draw a 1

            Returns
            -------
            interpretable_perturbed_samples
                Tensor of shape (nb_samples, num_features)
            """

            probs = tf.ones(num_features, tf.float32) * tf.cast(prob, tf.float32)
            uniform_sampling = tf.random.uniform(shape=[nb_samples, tf.squeeze(num_features)],
                                                dtype=tf.float32,
                                                minval=0,
                                                maxval=1)
            sample = tf.greater(probs, uniform_sampling)
            sample = tf.cast(sample, dtype=tf.int32)
            return sample

        return _default_pertub_function

    @staticmethod
    def _get_exp_kernel_func(
        distance_mode: str = "euclidean", kernel_width: float = 1.0
    ) -> Callable[[tf.Tensor, tf.Tensor, tf.Tensor], tf.Tensor]:
        """
        This method allow to get the function which compute:
            exp(-D(original_input,perturbed_sample)^2/kernel_width^2)
        Where D is the distance defined by distance mode.

        Parameters
        ----------
        distance_mode
            Can be either euclidian or cosine
        kernel_width
            The size of the kernel

        Returns
        -------
        similarity_kernel
            This callable should return distances between inputs and its perturbed samples
            (either in original space or in the interpretable space).
        """
        kernel_width = tf.cast(kernel_width,dtype=tf.float32)

        if distance_mode=="euclidean":
            @tf.function(
                input_signature = (
                    tf.TensorSpec(shape=None, dtype=tf.float32),
                    tf.TensorSpec(shape=[None, None], dtype=tf.int32),
                    tf.TensorSpec(shape=None, dtype=tf.float32)
                )
            )
            def _euclidean_similarity_kernel(
                original_input,
                interp_samples,
                perturbed_samples
            ) -> tf.Tensor:
            # pylint: disable=unused-argument

                augmented_input = tf.expand_dims(original_input, axis=0)
                augmented_input = tf.repeat(augmented_input, repeats=len(interp_samples), axis=0)

                flatten_inputs = tf.reshape(augmented_input, [len(interp_samples),-1])
                flatten_samples = tf.reshape(perturbed_samples, [len(interp_samples),-1])

                distances = tf.norm(flatten_inputs - flatten_samples, ord='euclidean', axis=1)

                similarities = tf.exp(-1.0 * (distances**2) / (kernel_width**2))

                return similarities

            return _euclidean_similarity_kernel

        if distance_mode=="cosine":
            @tf.function(
                input_signature = (
                    tf.TensorSpec(shape=None, dtype=tf.float32),
                    tf.TensorSpec(shape=[None, None], dtype=tf.int32),
                    tf.TensorSpec(shape=None, dtype=tf.float32)
                )
            )
            def _cosine_similarity_kernel(
                original_input,
                interp_samples,
                perturbed_samples
            ) -> tf.Tensor:
            # pylint: disable=unused-argument

                augmented_input = tf.expand_dims(original_input, axis=0)
                augmented_input = tf.repeat(augmented_input, repeats=len(interp_samples), axis=0)

                flatten_inputs = tf.reshape(augmented_input, [len(interp_samples),-1])
                flatten_samples = tf.reshape(perturbed_samples, [len(interp_samples),-1])

                distances = 1.0 - cosine_similarity(flatten_inputs, flatten_samples, axis=1)
                similarities = tf.exp(-1.0 * (distances**2) / (kernel_width**2))

                return similarities

            return _cosine_similarity_kernel

        raise ValueError("distance_mode must be either cosine or euclidean.")

    @staticmethod
    def _default_image_map_to_interpret_space(inp: tf.Tensor) -> tf.Tensor:
        """
        This method compute the quickshift segmentation.

        Parameters
        ----------
        inputs
            A single Input sample

        Returns
        -------
        mappings
            Mappings which map each pixel to the corresponding segment
        """
        mapping = quickshift(inp.numpy().astype('double'), ratio=0.5, kernel_size=2)
        mapping = tf.cast(mapping, tf.int32)

        return mapping

    @staticmethod
    def _default_2dimage_map_to_interpret_space(inp: tf.Tensor) -> tf.Tensor:
        """
        This method compute the felzenszwalb segmentation.

        Parameters
        ----------
        inputs
            A single Input sample

        Returns
        -------
        mappings
            Mappings which map each pixel to the corresponding segment
        """
        mapping = felzenszwalb(inp.numpy().astype('double'))
        mapping = tf.cast(mapping, tf.int32)

        return mapping


    @staticmethod
    def _default_tab_map_to_interpret_space(inp: tf.Tensor) -> tf.Tensor:
        """
        This method compute a similarity mapping i.e each features is independent.

        Parameters
        ----------
        inp
            A single Input sample

        Returns
        -------
        mappings
            Mappings which map each pixel to the corresponding segment
        """
        mapping = tf.range(len(inp))

        return mapping



"""
Sampling methods
"""

from abc import ABC, abstractmethod

import numpy as np
import tensorflow as tf
import scipy


class Sampler(ABC):
    """
    Base class for sampling.
    """

    def __init__(self, binary=False):
        self.binary = binary

    @abstractmethod
    def __call__(self, dimension, nb_design) -> np.ndarray:
        raise NotImplementedError()

    @staticmethod
    def make_binary(masks):
        """
        Transform [0, 1]^d Masks into binary {0, 1}^d masks.

        Parameters
        ----------
        masks
            Low resolution masks (before upsampling).

        Returns
        -------
        masks
            Binary masks.
        """
        return np.round(masks)


class ScipySampler(Sampler):
    """
    Base class based on Scipy qmc module for sampling.
    """

    def __init__(self, binary=False):
        super().__init__(binary)

        try:
            self.qmc = scipy.stats.qmc  # pylint: disable=E1101
        except AttributeError as err:
            raise ModuleNotFoundError(
                "Xplique need scipy>=1.7 to use this sampling."
            ) from err


class TFSobolSequence(Sampler):
    """
    Tensorflow Sobol LP tau sequence sampler.

    Ref. I. M. Sobol., The distribution of points in a cube and the accurate evaluation of
    integrals (1967).
    https://www.sciencedirect.com/science/article/abs/pii/0041555367901449
    """

    def __call__(self, dimension, nb_design):

        points = tf.math.sobol_sample(dimension, nb_design, dtype=tf.float32).numpy()
        if self.binary:
            points = self.make_binary(points)

        return points


class ScipySobolSequence(ScipySampler):
    """
    Scipy Sobol LP tau sequence sampler.

    Ref. I. M. Sobol., The distribution of points in a cube and the accurate evaluation of
    integrals (1967).
    https://www.sciencedirect.com/science/article/abs/pii/0041555367901449
    """

    def __call__(self, dimension, nb_design):
        sampler = self.qmc.Sobol(dimension, scramble=False)
        points = sampler.random(nb_design).astype(np.float32)
        if self.binary:
            points = self.make_binary(points)

        return points


class HaltonSequence(ScipySampler):
    """
    Halton sequence sampler.

    Ref. J.H. Halton., On the efficiency of certain quasi-random sequences of points in evaluating
    multi-dimensional integrals (1960).
    https://link.springer.com/article/10.1007/BF01386213
    """

    def __call__(self, dimension, nb_design):
        sampler = self.qmc.Halton(dimension, scramble=False)
        points = sampler.random(nb_design).astype(np.float32)
        if self.binary:
            points = self.make_binary(points)

        return points


class LatinHypercube(ScipySampler):
    """
    Latin hypercube replicated sampler.

    Ref. Mckay & al., A Comparison of Three Methods for Selecting Values of Input Variables in the
    Analysis of Output from a Computer Code (1979).
    https://www.jstor.org/stable/1268522
    """

    def __call__(self, dimension, nb_design):
        sampler = self.qmc.LatinHypercube(dimension)
        points = sampler.random(nb_design).astype(np.float32)
        if self.binary:
            points = self.make_binary(points)

        return points



import matplotlib
import numpy as np
import os

from xplique.plots.timeseries import plot_attributions


def test_one_explanation():
    # test timeseries plot attribution method with one explanation
    nb_features = 7
    nb_time_steps = 26
    title = "test one explanation"
    filepath = "tests/plots/timeseries_one_attribution_test.png"
    cmap = "coolwarm"

    features = ["feature_" + str(i) for i in range(nb_features)]
    explanations = np.random.uniform(size=(nb_time_steps, nb_features))

    plot_attributions(
        explanations,
        features,
        title,
        filepath,
        cmap=cmap,
        colorbar=True,
    )

    assert os.path.exists(filepath)
    os.remove(filepath)


def test_several_explanations():
    # test timeseries plot attribution method with several explanations
    nb_features = 7
    nb_time_steps = 26
    title = "test several explanations"
    filepath = "tests/plots/timeseries_several_attributions_test.png"
    cmap = "coolwarm"

    features = ["feature_" + str(i) for i in range(nb_features)]
    explanations = {
        "method_" + str(i):
            np.random.uniform(size=(nb_time_steps, nb_features))
        for i in range(10)
    }

    plot_attributions(
        explanations,
        features,
        title,
        filepath,
        cmap=cmap,
        colorbar=True,
    )

    assert os.path.exists(filepath)
    os.remove(filepath)



"""
Sobol Attribution Perturbation functions
"""
#pylint: disable=C0103,E1101

import cv2
import numpy as np
import tensorflow as tf

from ...types import Callable, Optional


@tf.function
def _baseline_ponderation(x, masks, x0):
    return tf.expand_dims(x, 0) * masks + (1.0 - masks) * tf.expand_dims(x0, 0)


@tf.function
def _amplitude_operator(x, masks, sigma):
    return x[None, :, :, :] * (masks - 0.5) * sigma


def inpainting(x: tf.Tensor) -> Callable:
    """
    Tensorflow inpainting perturbation function.

    X_perturbed = X * M

    Parameters
    ----------
    input
        Image to perform perturbation on.

    Returns
    -------
    inpainting_operator
        Inpainting perturbation function.
    """
    x0 = np.zeros(x.shape)
    x0 = tf.cast(x0, tf.float32)

    def f(masks):
        return _baseline_ponderation(x, masks, x0)
    return f


def blurring(x : tf.Tensor, sigma : Optional[int] = 10) -> Callable:
    """
    Tensorflow blur perturbation function.

    X_perturbed = blur(X, M)

    Parameters
    ----------
    input
        Image to perform perturbation on.
    sigma
        Blurring operator intensity.

    Returns
    -------
    blurring_operator
        Blur perturbation function.
    """
    x0 = cv2.blur(np.array(x, copy=True), (sigma, sigma))
    x0 = tf.cast(x0, tf.float32)

    if tf.rank(x0) == 2:
        x0 = x0[:, :, None]

    def f(masks):
        return _baseline_ponderation(x, masks, x0)
    return f


def amplitude(x : tf.Tensor, sigma : int = 1.0) -> Callable:
    """
    Tensorflow amplitude perturbation function.

    X_perturbed = X + (M - 0.5) * sigma

    Parameters
    ----------
    input
        Image to perform perturbation on.
    sigma
        Amplitude operator intensity.

    Returns
    -------
    ampitude_operator
        Blur perturbation function.
    """

    def f(masks):
        return _amplitude_operator(x, masks, sigma)
    return f



"""
Module related to VarGrad method
"""

import tensorflow as tf

from .smoothgrad import SmoothGrad


class VarGrad(SmoothGrad):
    """
     VarGrad is a variance analog to SmoothGrad.

    Ref. Adebayo & al., Sanity check for Saliency map (2018).
    https://papers.nips.cc/paper/8160-sanity-checks-for-saliency-maps.pdf

    Parameters
    ----------
    model
        The model from which we want to obtain explanations
    output_layer
        Layer to target for the outputs (e.g logits or after softmax).
        If an `int` is provided it will be interpreted as a layer index.
        If a `string` is provided it will look for the layer name.

        Default to the last layer.
        It is recommended to use the layer before Softmax.
    batch_size
        Number of inputs to explain at once, if None compute all at once.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].
    nb_samples
        Number of noisy samples generated for the smoothing procedure.
    noise
        Scalar, noise used as standard deviation of a normal law centered on zero.
    """

    @staticmethod
    @tf.function
    def _reduce_gradients(gradients: tf.Tensor) -> tf.Tensor:
        """
        Reduce the gradients using the variance obtained on each noisy samples.

        Parameters
        ----------
        gradients
            Gradients to reduce the sampling dimension for each inputs.

        Returns
        -------
        reduced_gradients
            Single saliency map for each input.
        """
        return tf.math.reduce_variance(gradients, axis=1)



import tensorflow as tf
import numpy as np

from xplique.features_visualizations.regularizers import (l1_reg, l2_reg, l_inf_reg,
                                                          total_variation_reg)
from ..utils import almost_equal


def test_lp_norm():
    vec = np.array([[[-4.0, 4.0]]])[np.newaxis, :]
    l1 = l1_reg(1.0)
    l2 = l2_reg(2.0)
    linf = l_inf_reg(10.0)

    assert almost_equal(l1(vec)[0], 4.0)
    assert almost_equal(l2(vec)[0], 8.0)
    assert almost_equal(linf(vec)[0], 40.0)


def test_tv():
    vec = tf.cast([[[0., 0., 0.],
                    [0., 2., 0.],
                    [0., 0., 1.]]], tf.float32)[:, :, :, tf.newaxis]
    tv = total_variation_reg(1.0)

    assert almost_equal(tv(vec)[0], 10.0)



"""
GSA Attribution Method base explainer
"""

from enum import Enum

import cv2
import numpy as np
import tensorflow as tf

from ...types import Callable, Union, Optional, Tuple, OperatorSignature
from ...commons import batch_tensor, repeat_labels, Tasks
from ..base import BlackBoxExplainer, sanitize_input_output
from .perturbations import amplitude, inpainting, blurring


class PerturbationFunction(Enum):
    """
    GSA Perturbation function interface.
    """

    INPAINTING = inpainting
    BLURRING = blurring
    AMPLITUDE = amplitude

    @staticmethod
    def from_string(perturbation_function: str) -> "PerturbationFunction":
        """
        Restore a perturbation function from a string.

        Parameters
        ----------
        perturbation_function
            String indicating the perturbation function to restore: must be one
            of 'inpainting', 'blurring' or 'amplitude'.

        Returns
        -------
        perturbation_function
            The PerturbationFunction object.
        """
        assert perturbation_function in [
            "inpainting",
            "blurring",
            "amplitude",
        ], "Only 'inpainting', 'blurring' and 'amplitude' are supported."

        if perturbation_function == "amplitude":
            return PerturbationFunction.AMPLITUDE
        if perturbation_function == "blurring":
            return PerturbationFunction.BLURRING
        return PerturbationFunction.INPAINTING


class GSABaseAttributionMethod(BlackBoxExplainer):
    """
    GSA base Attribution Method.
    Base explainer for all the attribution method based on Global Sensitivity Analysis.

    Parameters
    ----------
    model
        Model used for computing explanations.
    grid_size
        Cut the image in a grid of (grid_size, grid_size) to estimate an indice per cell.
    nb_design
        Must be a power of two. Number of design, the number of forward
        will be: nb_design * (grid_size**2 + 2). Generally not above 32.
    sampler
        Sampler function to call to generate masks.
    estimator
        Estimator used to compute the attribution score, e.g Sobol or HSIC estimator.
    perturbation_function
        Function to call to apply the perturbation on the input. Can also be string in
        'inpainting', 'blur'.
    batch_size
        Batch size to use for the forwards.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].
    """

    def __init__(
        self,
        model: tf.keras.Model,
        sampler: Callable,
        estimator: Callable,
        grid_size: int = 7,
        nb_design: int = 32,
        perturbation_function: Optional[Union[Callable, str]] = "inpainting",
        batch_size=256,
        operator: Optional[Union[Tasks, str, OperatorSignature]] = None,
    ):

        super().__init__(model, batch_size, operator)

        self.grid_size = grid_size
        self.nb_design = nb_design

        if isinstance(perturbation_function, str):
            self.perturbation_function = PerturbationFunction.from_string(
                perturbation_function
            )
        else:
            self.perturbation_function = perturbation_function

        self.sampler = sampler
        self.estimator = estimator

        self.masks = self.sampler(grid_size**2, nb_design).reshape(
            (-1, grid_size, grid_size, 1)
        )

    @sanitize_input_output
    def explain(
        self,
        inputs: Union[tf.data.Dataset, tf.Tensor, np.ndarray],
        targets: Optional[Union[tf.Tensor, np.ndarray]] = None,
    ) -> tf.Tensor:
        """
        Compute the total Sobol' indices according to the explainer parameter (perturbation
        function, grid size...). Accept Tensor, numpy array or tf.data.Dataset (in that case
        targets is None).

        Parameters
        ----------
        inputs
            Images to be explained, either tf.dataset, Tensor or numpy array.
            If Dataset, targets should not be provided (included in Dataset).
            Expected shape (N, W, H, C) or (N, W, H).
        targets
            One-hot encoding for classification or direction {-1, +1} for regression.
            Tensor or numpy array.
            Expected shape (N, C) or (N).

        Returns
        -------
        attributions_maps
            GSA Attribution Method explanations, same shape as the inputs except for the channels.
        """
        # pylint: disable=E1101

        input_shape = (inputs.shape[1], inputs.shape[2])
        heatmaps = None

        for inp, target in zip(inputs, targets):

            perturbator = self.perturbation_function(inp)
            outputs = None

            for batch_masks in batch_tensor(self.masks, self.batch_size):

                batch_x, batch_y = self._batch_perturbations(
                    batch_masks, perturbator, target, input_shape
                )
                batch_outputs = self.inference_function(self.model, batch_x, batch_y)

                outputs = (
                    batch_outputs
                    if outputs is None
                    else tf.concat([outputs, batch_outputs], axis=0)
                )

            heatmap = self.estimator(self.masks, outputs, self.nb_design)
            heatmap = cv2.resize(heatmap, input_shape, interpolation=cv2.INTER_CUBIC)[
                None, :, :
            ]

            heatmaps = (
                heatmap if heatmaps is None else tf.concat([heatmaps, heatmap], axis=0)
            )

        return heatmaps

    @staticmethod
    @tf.function
    def _batch_perturbations(
        masks: tf.Tensor,
        perturbator: Callable,
        target: tf.Tensor,
        input_shape: Tuple[int, int],
    ) -> Union[tf.Tensor, tf.Tensor]:
        """
        Prepare perturbated input and replicated targets before a batch inference.

        Parameters
        ----------
        masks
            Perturbation masks in lower dimensions (grid_size, grid_size).
        perturbator
            Perturbation function to be called with the upsampled masks.
        target
            Label of a single prediction
        input_shape
            Shape of a single input

        Returns
        -------
        perturbated_inputs
            One inputs perturbated for each masks, according to the pertubation function
            modulated by the masks values.
        repeated_targets
            Replicated labels, one for each masks.
        """
        repeated_targets = repeat_labels(target[None, :], len(masks))

        upsampled_masks = tf.image.resize(masks, input_shape, method="nearest")
        perturbated_inputs = perturbator(upsampled_masks)

        return perturbated_inputs, repeated_targets



"""
Fidelity (or Faithfulness) metrics
"""

from inspect import isfunction

import numpy as np
import tensorflow as tf
from scipy.stats import spearmanr

from .base import ExplanationMetric
from ..types import Union, Callable, Optional, Dict


class MuFidelity(ExplanationMetric):
    """
    Used to compute the fidelity correlation metric. This metric ensure there is a correlation
    between a random subset of pixels and their attribution score. For each random subset
    created, we set the pixels of the subset at a baseline state and obtain the prediction score.
    This metric measures the correlation between the drop in the score and the importance of the
    explanation.

    Ref. Bhatt & al., Evaluating and Aggregating Feature-based Model Explanations (2020).
    https://arxiv.org/abs/2005.00631 (def. 3)

    Notes
    -----
    As noted in the original article, the default operation selects pixel-wise subsets
    independently. However, when using medium or high dimensional images, it is recommended to
    select super-pixels, see the grid_size parameter.

    Parameters
    ----------
    model
        Model used for computing metric.
    inputs
        Input samples under study.
    targets
        One-hot encoded labels or regression target (e.g {+1, -1}), one for each sample.
    batch_size
        Number of samples to explain at once, if None compute all at once.
    grid_size
        If none, compute the original metric, else cut the image in (grid_size, grid_size) and
        each element of the subset will be a super pixel representing one element of the grid.
        You should use this when dealing with medium / large size images.
    subset_percent
        Percent of the image that will be set to baseline.
    baseline_mode
        Value of the baseline state, will be called with the a single input if it is a function.
    nb_samples
        Number of different subsets to try on each input to measure the correlation.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].
    activation
        A string that belongs to [None, 'sigmoid', 'softmax']. Specify if we should add
        an activation layer once the model has been called. It is useful, for instance
        if you want to measure a 'drop of probability' by adding a sigmoid or softmax
        after getting your logits. If None does not add a layer to your model.
    """

    def __init__(self,
                 model: Callable,
                 inputs: Union[tf.data.Dataset, tf.Tensor, np.ndarray],
                 targets: Optional[Union[tf.Tensor, np.ndarray]] = None,
                 batch_size: Optional[int] = 64,
                 grid_size: Optional[int] = 9,
                 subset_percent: float = 0.2,
                 baseline_mode: Union[Callable, float] = 0.0,
                 nb_samples: int = 200,
                 operator: Optional[Callable] = None,
                 activation: Optional[str] = None):
        # pylint: disable=R0913
        super().__init__(model, inputs, targets, batch_size, operator, activation)
        self.grid_size = grid_size
        self.subset_percent = subset_percent
        self.baseline_mode = baseline_mode
        self.nb_samples = nb_samples

        # if unspecified use the original equation (pixel-wise modification)
        self.grid_size = grid_size or inputs.shape[1]
        # cardinal of subset (|S| in the equation)
        self.subset_size = int(self.grid_size ** 2 * self.subset_percent)

        # prepare the random masks that will designate the modified subset (S in original equation)
        # we ensure the masks have exactly `subset_size` pixels set to baseline
        subset_masks = np.random.rand(self.nb_samples, self.grid_size ** 2)
        subset_masks = subset_masks.argsort(axis=-1) > self.subset_size

        # and interpolate them if needed
        subset_masks = subset_masks.astype(np.float32).reshape(
            (self.nb_samples, self.grid_size, self.grid_size, 1))
        self.subset_masks = tf.image.resize(subset_masks, inputs.shape[1:-1], method="nearest")

        self.base_predictions = self.batch_inference_function(self.model, inputs,
                                                              targets, self.batch_size)

    def evaluate(self,
                 explanations: Union[tf.Tensor, np.ndarray]) -> float:
        """
        Evaluate the fidelity score.

        Parameters
        ----------
        explanations
            Explanation for the inputs, labels to evaluate.

        Returns
        -------
        fidelity_score
            Metric score, average correlation between the drop in score when variables are set
            to a baseline state and the importance of these variables according to the
            explanations.
        """
        explanations = np.array(explanations)
        assert len(explanations) == len(self.inputs), "The number of explanations must be the " \
                                            f"same as the number of inputs: {len(explanations)}"\
                                            f" vs {len(self.inputs)}"

        correlations = []
        for inp, label, phi, base in zip(self.inputs, self.targets, explanations,
                                         self.base_predictions):
            label = tf.repeat(label[None, :], self.nb_samples, 0)
            baseline = self.baseline_mode(inp) if isfunction(self.baseline_mode) else \
                self.baseline_mode
            # use the masks to set the selected subsets to baseline state
            degraded_inputs = inp * self.subset_masks + (1.0 - self.subset_masks) * baseline
            # measure the two terms that should be correlated
            preds = base - self.batch_inference_function(self.model, degraded_inputs,
                                                         label, self.batch_size)

            attrs = tf.reduce_sum(phi * (1.0 - self.subset_masks), (1, 2, 3))
            corr_score = spearmanr(preds, attrs)[0]

            # sanity check: if the model predictions are the same, no variation
            if np.isnan(corr_score):
                corr_score = 0.0

            correlations.append(corr_score)

        fidelity_score = np.mean(correlations)

        return float(fidelity_score)


class CausalFidelity(ExplanationMetric):
    """
    Used to compute the insertion and deletion metrics.

    Parameters
    ----------
    model
        Model used for computing metric.
    inputs
        Input samples under study.
    targets
        One-hot encoded labels or regression target (e.g {+1, -1}), one for each sample.
    batch_size
        Number of samples to explain at once, if None compute all at once.
    causal_mode
        If 'insertion', the path is baseline to original image, for 'deletion' the path is original
        image to baseline.
    baseline_mode
        Value of the baseline state, will be called with the inputs if it is a function.
    steps
        Number of steps between the start and the end state.
        Can be set to -1 for all possible steps to be computed.
    max_percentage_perturbed
        Maximum percentage of the input perturbed.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].
    activation
        A string that belongs to [None, 'sigmoid', 'softmax']. Specify if we should add
        an activation layer once the model has been called. It is useful, for instance
        if you want to measure a 'drop of probability' by adding a sigmoid or softmax
        after getting your logits. If None does not add a layer to your model.
    """

    def __init__(self,
                 model: tf.keras.Model,
                 inputs: Union[tf.data.Dataset, tf.Tensor, np.ndarray],
                 targets: Optional[Union[tf.Tensor, np.ndarray]] = None,
                 batch_size: Optional[int] = 64,
                 causal_mode: str = "deletion",
                 baseline_mode: Union[float, Callable] = 0.0,
                 steps: int = 10,
                 max_percentage_perturbed: float = 1.0,
                 operator: Optional[Callable] = None,
                 activation: Optional[str] = None
                 ):
        # pylint: disable=R0913
        super().__init__(model, inputs, targets, batch_size, operator, activation)
        self.causal_mode = causal_mode
        self.baseline_mode = baseline_mode

        # If the input has channels (colored image), they are all occluded at the same time
        self.has_channels = len(inputs.shape) > 3

        if self.has_channels:
            self.nb_features = np.prod(inputs.shape[1:-1])
            self.inputs_flatten = inputs.reshape((len(inputs), self.nb_features, inputs.shape[-1]))
        else:
            self.nb_features = np.prod(inputs.shape[1:])
            self.inputs_flatten = inputs.reshape((len(inputs), self.nb_features, 1))

        assert 0.0 < max_percentage_perturbed <= 1.0, \
            "`max_percentage_perturbed` must be in ]0, 1]."
        self.max_nb_perturbed = tf.math.floor(self.nb_features * max_percentage_perturbed)

        if steps == -1:
            steps = self.max_nb_perturbed
        self.steps = steps

    def evaluate(self,
                 explanations: Union[tf.Tensor, np.ndarray]) -> float:
        """
        Evaluate the causal score.

        Parameters
        ----------
        explanations
            Explanation for the inputs, labels to evaluate.

        Returns
        -------
        causal_score
            Metric score, area over the deletion (lower is better) or insertion (higher is
            better) curve.
        """
        scores_dict = self.detailed_evaluate(explanations)

        # compute auc using trapezoidal rule (the steps are equally distributed)
        np_scores = np.array(list(scores_dict.values()))
        auc = np.mean(np_scores[:-1] + np_scores[1:]) * 0.5

        return auc

    def detailed_evaluate(self,
                          explanations: Union[tf.Tensor, np.ndarray]) -> Dict[int, float]:
        """
        Evaluate model performance for successive perturbations of an input.
        Used to compute causal score.

        The successive perturbations in the Insertion and Deletion metrics create a list of scores.
        This list of scores make a score evolution curve.
        The AUC of such curve is used as an explanation metric.
        However, the curve in itself is rich in information,
        its visualization and interpretation can bring further comprehension
        on the explanation and the model.
        Therefore this method was added so that it is possible to construct such curves.

        Parameters
        ----------
        explanations
            Explanation for the inputs, labels to evaluate.

        Returns
        -------
        causal_score_dict
            Dictionary of scores obtain for different perturbations
            Keys are the steps, i.e the number of features perturbed
            Values are the scores, the score of the model
                on the inputs with the corresponding number of features perturbed
        """
        explanations = np.array(explanations)
        assert len(explanations) == len(self.inputs), "The number of explanations must be the " \
                                            f"same as the number of inputs: {len(explanations)}"\
                                            f"vs {len(self.inputs)}"

        # the reference does not specify how to manage the channels of the explanations
        if len(explanations.shape) == 4:
            explanations = np.mean(explanations, -1)

        explanations_flatten = explanations.reshape((len(explanations), -1))

        # for each sample, sort by most important features according to the explanation
        most_important_features = np.argsort(explanations_flatten, axis=-1)[:, ::-1]

        baselines = self.baseline_mode(self.inputs) if isfunction(self.baseline_mode) else \
            np.ones_like(self.inputs, dtype=np.float32) * self.baseline_mode
        baselines_flatten = baselines.reshape(self.inputs_flatten.shape)

        steps = np.linspace(0, self.max_nb_perturbed, self.steps + 1, dtype=np.int32)

        if self.causal_mode == "deletion":
            start = self.inputs_flatten
            end = baselines_flatten
        elif self.causal_mode == "insertion":
            start = baselines_flatten
            end = self.inputs_flatten
        else:
            raise NotImplementedError(f'Unknown causal mode `{self.causal_mode}`.')

        scores_dict = {}
        for step in steps:
            ids_to_flip = most_important_features[:, :step]
            batch_inputs = start.copy()

            for i, ids in enumerate(ids_to_flip):
                batch_inputs[i, ids] = end[i, ids]

            batch_inputs = batch_inputs.reshape((-1, *self.inputs.shape[1:]))

            predictions = self.batch_inference_function(self.model, batch_inputs,
                                                        self.targets, self.batch_size)

            scores_dict[step] = np.mean(predictions)

        return scores_dict


class Deletion(CausalFidelity):
    """
    The deletion metric measures the drop in the probability of a class as important pixels (given
    by the saliency map) are gradually removed from the image. A sharp drop, and thus a small
    area under the probability curve, are indicative of a good explanation.

    Ref. Petsiuk & al., RISE: Randomized Input Sampling for Explanation of Black-box Models (2018).
    https://arxiv.org/pdf/1806.07421.pdf

    Parameters
    ----------
    model
        Model used for computing metric.
    inputs
        Input samples under study.
    targets
        One-hot encoded labels or regression target (e.g {+1, -1}), one for each sample.
    batch_size
        Number of samples to explain at once, if None compute all at once.
    baseline_mode
        Value of the baseline state, will be called with the inputs if it is a function.
    steps
        Number of steps between the start and the end state.
        Can be set to -1 for all possible steps to be computed.
    max_percentage_perturbed
        Maximum percentage of the input perturbed.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].
    activation
        A string that belongs to [None, 'sigmoid', 'softmax']. Specify if we should add
        an activation layer once the model has been called. It is useful, for instance
        if you want to measure a 'drop of probability' by adding a sigmoid or softmax
        after getting your logits. If None does not add a layer to your model.
    """

    def __init__(self,
                 model: tf.keras.Model,
                 inputs: Union[tf.data.Dataset, tf.Tensor, np.ndarray],
                 targets: Optional[Union[tf.Tensor, np.ndarray]] = None,
                 batch_size: Optional[int] = 64,
                 baseline_mode: Union[float, Callable] = 0.0,
                 steps: int = 10,
                 max_percentage_perturbed: float = 1.0,
                 operator: Optional[Callable] = None,
                 activation: Optional[str] = None
                 ):
        super().__init__(model, inputs, targets, batch_size, "deletion",
                         baseline_mode, steps, max_percentage_perturbed,
                         operator, activation)


class Insertion(CausalFidelity):
    """
    The insertion metric, on the other hand, captures the importance of the pixels in terms of
    their ability to synthesize an image and is measured by the rise in the probability of the
    class of interest as pixels are added according to the generated importance map.

    Ref. Petsiuk & al., RISE: Randomized Input Sampling for Explanation of Black-box Models (2018).
    https://arxiv.org/pdf/1806.07421.pdf

    Parameters
    ----------
    model
        Model used for computing metric.
    inputs
        Input samples under study.
    targets
        One-hot encoded labels or regression target (e.g {+1, -1}), one for each sample.
    batch_size
        Number of samples to explain at once, if None compute all at once.
    baseline_mode
        Value of the baseline state, will be called with the inputs if it is a function.
    steps
        Number of steps between the start and the end state.
        Can be set to -1 for all possible steps to be computed.
    max_percentage_perturbed
        Maximum percentage of the input perturbed.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].
    activation
        A string that belongs to [None, 'sigmoid', 'softmax']. Specify if we should add
        an activation layer once the model has been called. It is useful, for instance
        if you want to measure a 'drop of probability' by adding a sigmoid or softmax
        after getting your logits. If None does not add a layer to your model.
    """

    def __init__(self,
                 model: tf.keras.Model,
                 inputs: Union[tf.data.Dataset, tf.Tensor, np.ndarray],
                 targets: Optional[Union[tf.Tensor, np.ndarray]] = None,
                 batch_size: Optional[int] = 64,
                 baseline_mode: Union[float, Callable] = 0.0,
                 steps: int = 10,
                 max_percentage_perturbed: float = 1.0,
                 operator: Optional[Callable] = None,
                 activation: Optional[str] = None
                 ):
        super().__init__(model, inputs, targets, batch_size, "insertion",
                         baseline_mode, steps, max_percentage_perturbed,
                         operator, activation)



"""
Module related to RISE method
"""

import tensorflow as tf
import numpy as np

from .base import BlackBoxExplainer, sanitize_input_output
from ..commons import repeat_labels, batch_tensor, Tasks
from ..types import Callable, Optional, Union, Tuple, OperatorSignature


class Rise(BlackBoxExplainer):
    """
    Used to compute the RISE method, by probing the model with randomly masked versions of
    the input image and obtaining the corresponding outputs to deduce critical areas.

    Ref. Petsiuk & al., RISE: Randomized Input Sampling for Explanation of Black-box Models (2018).
    https://arxiv.org/abs/1806.07421

    Parameters
    ----------
    model
        The model from which we want to obtain explanations
    batch_size
        Number of pertubed samples to explain at once.
        Default to 32.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].
    nb_samples
        Number of masks generated for Monte Carlo sampling.
    grid_size
        Size of the grid used to generate the scaled-down masks. Masks are then rescale to
        and cropped to input_size. Can be a tuple for different cutting depending on the dimension.
    preservation_probability
        Probability of preservation for each pixel (or the percentage of non-masked pixels in
        each masks), also the expectation value of the mask.
    """

    # Avoid zero division during procedure. (the value is not important, as if the denominator is
    # zero, then the nominator will also be zero).
    EPSILON = tf.constant(1e-4)

    def __init__(self,
                 model: Callable,
                 batch_size: Optional[int] = 32,
                 operator: Optional[Union[Tasks, str, OperatorSignature]] = None,
                 nb_samples: int = 4000,
                 grid_size: Union[int, Tuple[int]] = 7,
                 preservation_probability: float = .5):
        super().__init__(model, batch_size, operator)

        self.nb_samples = nb_samples
        self.grid_size = grid_size
        self.preservation_probability = preservation_probability
        self.binary_masks = Rise._get_masks(self.nb_samples, self.grid_size,
                                            self.preservation_probability)

    @sanitize_input_output
    def explain(self,
                inputs: Union[tf.data.Dataset, tf.Tensor, np.ndarray],
                targets: Optional[Union[tf.Tensor, np.ndarray]] = None) -> tf.Tensor:
        """
        Compute RISE for a batch of samples.

        Parameters
        ----------
        inputs
            Dataset, Tensor or Array. Input samples to be explained.
            If Dataset, targets should not be provided (included in Dataset).
            Expected shape among (N, W), (N, T, W), (N, H, W, C).
            More information in the documentation.
        targets
            Tensor or Array. One-hot encoding of the model's output from which an explanation
            is desired. One encoding per input and only one output at a time. Therefore,
            the expected shape is (N, output_size).
            More information in the documentation.

        Returns
        -------
        explanations
            RISE maps, same shape as the inputs, except for the channels.
        """
        rise_maps = None
        batch_size = self.batch_size or self.nb_samples

        # since the number of masks is often very large, we process the entries one by one
        for single_input, single_target in zip(inputs, targets):

            rise_nominator   = tf.zeros((*single_input.shape[:-1], 1))
            rise_denominator = tf.zeros((*single_input.shape[:-1], 1))

            # we iterate on the binary masks since they are cheap in memory
            for batch_masks in batch_tensor(self.binary_masks, batch_size):
                # the upsampling/cropping phase is performed on the batched masks
                masked_inputs, masks_upsampled = Rise._apply_masks(single_input, batch_masks)
                repeated_targets = repeat_labels(single_target[tf.newaxis, :], len(batch_masks))

                predictions = self.inference_function(self.model, masked_inputs, repeated_targets)

                rise_nominator += tf.reduce_sum(tf.reshape(predictions, (-1, 1, 1, 1))
                                                * masks_upsampled, 0)
                rise_denominator += tf.reduce_sum(masks_upsampled, 0)

            rise_map = rise_nominator / (rise_denominator + Rise.EPSILON)
            rise_map = rise_map[tf.newaxis, :, :, 0]

            rise_maps = rise_map if rise_maps is None else tf.concat([rise_maps, rise_map], axis=0)

        return rise_maps


    @staticmethod
    @tf.function
    def _get_masks(nb_samples: int,
                   grid_size: Union[int, Tuple[int]],
                   preservation_probability: float) -> tf.Tensor:
        """
        Random mask generation.
        Start by generating random mask in a lower dimension. Then,a bilinear interpolation to
        upsample the masks and take a random crop of the size of the inputs.

        Parameters
        ----------
        input_shape
            Shape of an input sample.
        nb_samples
            Number of masks generated for Monte Carlo sampling.
        grid_size
            Size of the grid used to generate the scaled-down masks.
            Can be a tuple for non-square grid.
        preservation_probability
            Probability of preservation for each pixel (or the percentage of non-masked pixels in
            each masks), also the expectation value of the mask.

        Returns
        -------
        binary_masks
            The downsampled binary masks.
        """
        if not isinstance(grid_size, tuple):
            downsampled_shape = (grid_size, grid_size)
        else:
            downsampled_shape = grid_size
        downsampled_masks = tf.random.uniform((nb_samples, *downsampled_shape, 1), 0, 1)

        binary_masks = downsampled_masks < preservation_probability

        return binary_masks

    @staticmethod
    @tf.function
    def _apply_masks(
        single_input: tf.Tensor,
        binary_masks: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor]:
        """
        Given input samples and masks, apply it for every sample and repeat the labels.

        Parameters
        ----------
        current_input
            Input samples to be explained.
        binary_masks
            Binary downsampled masks randomly generated.

        Returns
        -------
        masked_input
            All the masked combinations of the input (for each masks).
        masks
            Masks after the upsampling / cropping operation
        """
        # the upsampled size is defined as (h+1)(H/h) = H(1 + 1 / h)
        upsampled_size = (int(single_input.shape[0] * (1.0 + 1.0 / binary_masks.shape[1])),
                          int(single_input.shape[1] * (1.0 + 1.0 / binary_masks.shape[2])),)
        upsampled_size = tf.cast(upsampled_size, tf.int32)

        upsampled_masks = tf.image.resize(tf.cast(binary_masks, tf.float32), upsampled_size)

        masks = tf.image.random_crop(upsampled_masks, (binary_masks.shape[0],
                                                       *single_input.shape[:-1], 1))

        masked_input = tf.expand_dims(single_input, 0) * masks

        return masked_input, masks



"""
Module related to Grad-CAM method
"""

import tensorflow as tf
import numpy as np
import cv2

from .base import WhiteBoxExplainer, sanitize_input_output
from ..commons import find_layer, Tasks
from ..types import Tuple, Union, Optional, OperatorSignature


class GradCAM(WhiteBoxExplainer):
    """
    Used to compute the Grad-CAM visualization method.

    Only for Convolutional Networks.

    Ref. Selvaraju & al., Grad-CAM: Visual Explanations from Deep Networks via Gradient-based
    Localization (2016).
    https://arxiv.org/abs/1610.02391

    Parameters
    ----------
    model
        The model from which we want to obtain explanations
    output_layer
        Layer to target for the outputs (e.g logits or after softmax).
        If an `int` is provided it will be interpreted as a layer index.
        If a `string` is provided it will look for the layer name.

        Default to the last layer.
        It is recommended to use the layer before Softmax.
    batch_size
        Number of inputs to explain at once, if None compute all at once.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].
    conv_layer
        Layer to target for Grad-CAM algorithm.
        If an int is provided it will be interpreted as a layer index.
        If a string is provided it will look for the layer name.
    """

    def __init__(self,
                 model: tf.keras.Model,
                 output_layer: Optional[Union[str, int]] = None,
                 batch_size: Optional[int] = 32,
                 operator: Optional[Union[Tasks, str, OperatorSignature]] = None,
                 conv_layer: Optional[Union[str, int]] = None):
        super().__init__(model, output_layer, batch_size, operator)

        # find the layer to apply grad-cam
        if conv_layer is not None:
            self.conv_layer = find_layer(model, conv_layer)
        else:
            # no conv_layer specified, assuming default procedure : the last conv layer
            self.conv_layer = next(
                layer for layer in model.layers[::-1] if hasattr(layer, 'filters'))

        self.model = tf.keras.Model(model.input, [self.conv_layer.output, self.model.output])

    @sanitize_input_output
    def explain(self,
                inputs: Union[tf.data.Dataset, tf.Tensor, np.ndarray],
                targets: Optional[Union[tf.Tensor, np.ndarray]] = None) -> tf.Tensor:
        """
        Compute and resize explanations to match inputs shape.
        Accept Tensor, numpy array or tf.data.Dataset (in that case targets is None)

        Parameters
        ----------
        inputs
            Dataset, Tensor or Array. Input samples to be explained.
            If Dataset, targets should not be provided (included in Dataset).
            Expected shape among (N, W), (N, T, W), (N, H, W, C).
            More information in the documentation.
        targets
            Tensor or Array. One-hot encoding of the model's output from which an explanation
            is desired. One encoding per input and only one output at a time. Therefore,
            the expected shape is (N, output_size).
            More information in the documentation.

        Returns
        -------
        grad_cam
            Grad-CAM explanations, same shape as the inputs except for the channels.
        """
        # pylint: disable=E1101
        grad_cams = None
        batch_size = self.batch_size if self.batch_size is not None else len(inputs)

        for x_batch, y_batch in tf.data.Dataset.from_tensor_slices((inputs, targets)).batch(
                batch_size):
            batch_feature_maps, batch_gradients = GradCAM._gradient(self.model, x_batch, y_batch)
            batch_weights = self._compute_weights(batch_gradients, batch_feature_maps)
            batch_grad_cams = GradCAM._apply_weights(batch_weights, batch_feature_maps)

            grad_cams = batch_grad_cams if grad_cams is None else tf.concat(
                [grad_cams, batch_grad_cams], axis=0)

        # as Grad-CAM is based on the last convolutionnal layer, the explanation output has the
        # same dimensions as this layer, we need to resize the size of the explanations to match
        # the size of the inputs
        input_shape = inputs.shape[1:3]
        grad_cams = np.array([
            cv2.resize(grad_cam, (*input_shape,), interpolation=cv2.INTER_CUBIC)
            for grad_cam in grad_cams.numpy()])

        return grad_cams

    @staticmethod
    @tf.function
    def _gradient(model: tf.keras.Model,
                  inputs: tf.Tensor,
                  targets: tf.Tensor) -> Tuple[tf.Tensor,
                                              tf.Tensor]:
        """
        Compute the gradient with respect to the conv_layer.

        Parameters
        ----------
        model
            The model from which we want to obtain explanations.
        inputs
            Dataset, Tensor or Array. Input samples to be explained.
            If Dataset, targets should not be provided (included in Dataset).
            Expected shape among (N, W), (N, T, W), (N, H, W, C). More information in the
            documentation (API Description).
        targets
            Tensor or Array. One-hot encoding of the model's output from which an explanation
            is desired. One encoding per input and only one output at a time.
            More information in the documentation (API Description).

        Returns
        -------
        feature_maps
            Activations for the target convolution layer.
        feature_maps_gradients
            Gradients for the target convolution layer.
        """
        with tf.GradientTape(watch_accessed_variables=False) as tape:
            tape.watch(inputs)
            feature_maps, predictions = model(inputs)
            score = tf.reduce_sum(tf.multiply(predictions, targets), axis=-1)

        feature_maps_gradients = tape.gradient(score, feature_maps)

        return feature_maps, feature_maps_gradients

    @staticmethod
    @tf.function
    def _compute_weights(feature_maps_gradients: tf.Tensor,
                         feature_maps: tf.Tensor) -> tf.Tensor:
        """
        Compute the weights according to Grad-CAM procedure.

        Parameters
        ----------
        feature_maps_gradients
            Gradients for the target convolution layer.
        feature_maps
            Activations for the target convolution layer. Not used for Grad-CAM.

        Returns
        -------
        weights
            Weights for each feature maps.
        """
        # pylint: disable=unused-argument
        weights = tf.reduce_mean(feature_maps_gradients, axis=(1, 2), keepdims=True)
        return weights

    @staticmethod
    @tf.function
    def _apply_weights(weights: tf.Tensor,
                       feature_maps: tf.Tensor) -> tf.Tensor:
        """
        Apply the weights to the feature maps, sum them and follow it by a ReLU.

        Parameters
        ----------
        weights
            Weights for each feature maps.
        feature_maps
            Activations for the target convolution layer.

        Returns
        -------
        weighted_feature_maps
        """
        weighted_feature_maps = tf.reduce_sum(tf.multiply(feature_maps, weights), axis=-1)
        weighted_feature_maps = tf.nn.relu(weighted_feature_maps)

        return weighted_feature_maps



"""
Custom tensorflow operator for Attributions
"""

import inspect
from enum import Enum

import tensorflow as tf

from ..types import Callable, Optional, Union, OperatorSignature
from .exceptions import raise_invalid_operator, no_gradients_available
from .callable_operations import predictions_one_hot_callable

@tf.function
def predictions_operator(model: Callable,
                         inputs: tf.Tensor,
                         targets: tf.Tensor) -> tf.Tensor:
    """
    Compute predictions scores, only for the label class, for a batch of samples.

    Parameters
    ----------
    model
        Model used for computing predictions.
    inputs
        Input samples to be explained.
    targets
        One-hot encoded labels or regression target (e.g {+1, -1}), one for each sample.

    Returns
    -------
    scores
        Predictions scores computed, only for the label class.
    """
    scores = tf.reduce_sum(model(inputs) * targets, axis=-1)
    return scores

@tf.function
def regression_operator(model: Callable,
                        inputs: tf.Tensor,
                        targets: tf.Tensor) -> tf.Tensor:
    """
    Compute the the mean absolute error between model prediction and the target.
    Target should the model prediction on non-perturbed input.
    This operator can be used to compute attributions for all outputs of a regression model.

    Parameters
    ----------
    model
        Model used for computing predictions.
    inputs
        Input samples to be explained.
    targets
        Model prediction on non-perturbed inputs.

    Returns
    -------
    scores
        MAE between model prediction and targets.
    """
    scores = tf.reduce_mean(tf.abs(model(inputs) - targets), axis=-1)
    return scores


@tf.function
def binary_segmentation_operator(model: Callable,
                                 inputs: tf.Tensor,
                                 targets: tf.Tensor) -> tf.Tensor:
    """
    Compute the segmentation score for a batch of samples.

    Parameters
    ----------
    model
        Model used for computing predictions.
    inputs
        Input samples to be explained.
    targets
        One-hot encoded labels or regression target (e.g {+1, -1}), one for each sample.

    Returns
    -------
    scores
        Segmentation scores computed.
    """
    scores = tf.reduce_sum(model(inputs) * targets, axis=(1, 2))
    return scores


@tf.function
def segmentation_operator(model: Callable,
                          inputs: tf.Tensor,
                          targets: tf.Tensor) -> tf.Tensor:
    """
    Compute the segmentation score for a batch of samples.

    Parameters
    ----------
    model
        Model used for computing predictions.
    inputs
        Input samples to be explained.
    targets
        One-hot encoded labels or regression target (e.g {+1, -1}), one for each sample.

    Returns
    -------
    scores
        Segmentation scores computed.
    """
    scores = tf.reduce_sum(model(inputs) * targets, axis=(1, 2, 3))
    return scores

class Tasks(Enum):
    """
    Enumeration of different tasks for which we have defined operators
    """
    CLASSIFICATION = predictions_operator
    REGRESSION = regression_operator

    @staticmethod
    def from_string(operator_name: str) -> "Tasks":
        """
        Restore an operator from a string

        Parameters
        ----------
        operator_name
            String indicating the operator to restore: must be one
            of 'classification' or 'regression'

        Returns
        -------
        operator
            The Tasks object
        """
        assert operator_name in [
            "classification",
            "regression",
        ], "Only 'classification' and 'regression' are supported."

        if operator_name == "regression":
            return Tasks.REGRESSION
        return Tasks.CLASSIFICATION

def check_operator(operator: Callable):
    """
    Check if the operator is valid g(f, x, y) -> tf.Tensor
    and raise an exception and return true if so.

    Parameters
    ----------
    operator
        Operator to check

    Returns
    -------
    is_valid
        True if the operator is valid, False otherwise.
    """
    # handle tf functions
    # pylint: disable=protected-access
    if hasattr(operator, '_python_function'):
        return check_operator(operator._python_function)

    # the operator must be callable
    if not hasattr(operator, '__call__'):
        raise_invalid_operator()

    # the operator should take at least three arguments
    args = inspect.getfullargspec(operator).args
    if len(args) < 3:
        raise_invalid_operator()

    return True

def get_operator(
        operator: Optional[Union[Tasks, str, OperatorSignature]]):
    """
    This function allows to retrieve an operator from: a Tasks, a task name. If the operator
    is a custom one, we simply check if its signature is correct

    Parameters
    ----------
    operator
        An operator from the Tasks enum or the task name or a custom operator. If None, use a
        classification operator.

    Returns
    -------
    operator
        The operator requested
    """
    # case when no operator is provided
    if operator is None:
        return predictions_operator

    # case when the query is a string
    if isinstance(operator, str):
        return Tasks.from_string(operator)

    # case when the query belong to the Tasks enum
    if operator in [t.value for t in Tasks]:
        return operator

    # case when the operator is a custom one
    assert check_operator(operator)
    return operator

def get_gradient_of_operator(operator):
    """
    Get the gradient of an operator.

    Parameters
    ----------
    operator
        Operator to compute the gradient of.

    Returns
    -------
    gradient
        Gradient of the operator.
    """
    @tf.function
    def gradient(model, inputs, targets):
        with tf.GradientTape() as tape:
            tape.watch(inputs)
            scores = operator(model, inputs, targets)

        return tape.gradient(scores, inputs)

    return gradient


def operator_batching(operator: OperatorSignature) -> tf.Tensor:
    """
    Take care of batching an operator: (model, inputs, labels).

    Parameters
    ----------
    operator
        Any callable that take model, inputs and labels as parameters.

    Returns
    -------
    batched_operator
        Function that apply operator by batch.
    """

    def batched_operator(model, inputs, targets, batch_size=None):
        if batch_size is not None:
            dataset = tf.data.Dataset.from_tensor_slices((inputs, targets))
            results = tf.concat([
                operator(model, x, y)
                for x, y in dataset.batch(batch_size)
            ], axis=0)
        else:
            results = operator(model, inputs, targets)

        return results

    return batched_operator


batch_predictions = operator_batching(predictions_operator)
gradients_predictions = get_gradient_of_operator(predictions_operator)
batch_gradients_predictions = operator_batching(gradients_predictions)
batch_predictions_one_hot_callable = operator_batching(predictions_one_hot_callable)


def get_inference_function(
        model: Callable,
        operator: Optional[OperatorSignature] = None):
    """
    Define the inference function according to the model type

    Parameters
    ----------
    model
        Model used for computing explanations.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].

    Returns
    -------
    inference_function
        Same definition as the operator.
    batch_inference_function
        An inference function which treat inputs and targets by batch,
        it has an additionnal parameter `batch_size`.
    """
    if operator is not None:
        # user specified a string, an operator from the ones available or a
        # custom operator, we check if the operator is valid
        # and we wrap it to generate a batching version of this operator
        operator = get_operator(operator)
        inference_function = operator
        batch_inference_function = operator_batching(operator)

    elif isinstance(model, (tf.keras.Model, tf.Module, tf.keras.layers.Layer)):
        inference_function = predictions_operator
        batch_inference_function = batch_predictions

    else:
        # completely unknown model (e.g. sklearn), we can't backprop through it
        inference_function = predictions_one_hot_callable
        batch_inference_function = batch_predictions_one_hot_callable

    return inference_function, batch_inference_function


def get_gradient_functions(
        model: Callable,
        operator: Optional[OperatorSignature] = None):
    """
    Define the gradient function according to the model type

    Parameters
    ----------
    model
        Model used for computing explanations.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].

    Returns
    -------
    gradient
        Gradient function of the operator.
    batch_gradient
        An gradient function which treat inputs and targets by batch,
        it has an additionnal parameter `batch_size`.
    """
    if operator is not None:
        # user specified a string, an operator from the ones available or a
        # custom operator, we check if the operator is valid
        # and we wrap it to generate a batching version of this operator
        operator = get_operator(operator)
        gradient = get_gradient_of_operator(operator)
        batch_gradient = operator_batching(gradient)

    elif isinstance(model, tf.keras.Model):
        # no custom operator, for keras model we can backprop through the model
        gradient = gradients_predictions
        batch_gradient = batch_gradients_predictions

    else:
        # custom model or completely unknown model (e.g. sklearn), we can't backprop through it
        gradient = no_gradients_available
        batch_gradient = no_gradients_available

    return gradient, batch_gradient
    


"""
Utility classes and functions
"""

from .data_conversion import tensor_sanitize, numpy_sanitize
from .model_override import guided_relu_policy, deconv_relu_policy, override_relu_gradient, \
                            find_layer, open_relu_policy
from .tf_operations import repeat_labels, batch_tensor
from .callable_operations import predictions_one_hot_callable
from .operators import Tasks, get_operator, check_operator, operator_batching,\
                       get_inference_function, get_gradient_functions
from .exceptions import no_gradients_available, raise_invalid_operator
from .forgrad import forgrad



import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import Dense, Conv2D, Activation, Dropout, Flatten, MaxPooling2D, Input

from xplique.attributions import GradCAMPP
from ..utils import generate_data, almost_equal

def _generate_model(input_shape=(32, 32, 3), output_shape=10):
    model = tf.keras.Sequential()
    model.add(Input(shape=input_shape))
    model.add(Conv2D(4, kernel_size=(2, 2),
                     activation='relu', name='conv2d'))
    model.add(Conv2D(4, kernel_size=(2, 2),
                     activation='relu', name='conv2d_1'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Dropout(0.25))
    model.add(Flatten())
    model.add(Dense(output_shape))
    model.add(Activation('softmax'))
    model.compile(loss='categorical_crossentropy', optimizer='sgd')

    return model


def test_output_shape():
    """The output shape must be the same as the input shape"""

    input_shapes = [(28, 28, 1), (32, 32, 3)]
    nb_labels = 10

    for input_shape in input_shapes:
        samples, labels = generate_data(input_shape, nb_labels, 100)
        model = _generate_model(input_shape, nb_labels)

        method = GradCAMPP(model, -2)
        outputs = method.explain(samples, labels)

        assert samples.shape[:3] == outputs.shape[:3]


def test_conv_layer():
    """We should target the right layer using either int, string or default procedure"""
    tf.keras.backend.clear_session()

    model = _generate_model()

    last_conv_layer = model.get_layer('conv2d_1')
    first_conv_layer = model.get_layer('conv2d')
    flatten_layer = model.get_layer('flatten')

    # default should target the last conv layer
    gc_default = GradCAMPP(model)
    assert gc_default.conv_layer == last_conv_layer

    # target the first conv layer
    gc_input_conv = GradCAMPP(model, conv_layer=0)
    assert gc_input_conv.conv_layer == first_conv_layer

    # target a random flatten layer
    gc_flatten = GradCAMPP(model, conv_layer='flatten')
    assert gc_flatten.conv_layer == flatten_layer

def test_weights_computation():
    """Ensure the grad-cam++ weights are correct"""
    activations = np.array([
        [[1.0, 1.0],
         [1.0, 1.0]],

        [[1.0, 1.0],
         [0.0, 0.0]],

        [[0.0, 0.0],
         [0.0, 0.0]],

        [[0.5, 0.0],
         [0.0, 0.0]],
    ], dtype=np.float32)[None, :, :, :]
    grads = np.array([
        [[1.0, 1.0],
         [1.0, 1.0]],

        [[1.0, 1.0],
         [0.0, 0.0]],

        [[0.0, 0.0],
         [0.0, 0.0]],

        [[0.5, 0.0],
         [0.0, 0.0]],
    ], dtype=np.float32)[None, :, :, :]

    # move so that the filters F are at the end [F, W, H] -> [W, H, F]
    activations = np.moveaxis(activations, 1, 3)
    grads = np.moveaxis(grads, 1, 3)

    weights = GradCAMPP._compute_weights(grads, activations)
    assert almost_equal(weights[0], [4.0 / (2.0 * 4.0 + 4.0) * (4.0 / 4.0),
                                     2.0 / (2.0 * 2.0  + 1.0) * (2.0 / 4.0),
                                     0.0,
                                     0.25/ (2.0 * 0.25 + 0.5**3*0.125) * 0.125], 1e-3)

    grad_cam_pp = GradCAMPP._apply_weights(weights, activations)
    assert almost_equal(grad_cam_pp, np.sum(activations * weights, -1)) # as we have no negative value



"""
Image regularizers
"""

import tensorflow as tf

from ..types import Callable


def l1_reg(factor: float = 1.0) -> Callable:
    """
    Mean L1 regularization.

    Parameters
    ----------
    factor
        Weight that controls the importance of the regularization term.

    Returns
    -------
    reg
        Mean L1 of the images.
    """
    def reg(images: tf.Tensor) -> tf.Tensor:
        return factor * tf.reduce_mean(tf.abs(images), (1, 2, 3))
    return reg


def l2_reg(factor: float = 1.0) -> Callable:
    """
    Mean L2 regularization.

    Parameters
    ----------
    factor
        Weight that controls the importance of the regularization term.

    Returns
    -------
    reg
        Mean L2 of the images.
    """
    def reg(images: tf.Tensor) -> tf.Tensor:
        return factor * tf.sqrt(tf.reduce_mean(images**2, (1, 2, 3)))
    return reg


def l_inf_reg(factor: float = 1.0) -> Callable:
    """
    Mean L-inf regularization.

    Parameters
    ----------
    factor
        Weight that controls the importance of the regularization term.

    Returns
    -------
    l_inf
        Max of the images.
    """
    def l_inf(images: tf.Tensor) -> tf.Tensor:
        return factor * tf.reduce_max(tf.abs(images), (1, 2, 3))
    return l_inf


def total_variation_reg(factor: float = 1.0) -> Callable:
    """
    Total variation regularization.

    Parameters
    ----------
    factor
        Weight that controls the importance of the regularization term.

    Returns
    -------
    tv_reg
        Total variation of the images.
    """
    def tv_reg(images: tf.Tensor) -> tf.Tensor:
        return factor * tf.image.total_variation(images)
    return tv_reg



"""
Module related to SmoothGrad method
"""

import tensorflow as tf
import numpy as np

from .base import WhiteBoxExplainer, sanitize_input_output
from ..commons import repeat_labels, batch_tensor, Tasks
from ..types import Tuple, Union, Optional, OperatorSignature


class SmoothGrad(WhiteBoxExplainer):
    """
    Used to compute the SmoothGrad, by averaging Saliency maps of noisy samples centered on the
    original sample.

    Ref. Smilkov & al., SmoothGrad: removing noise by adding noise (2017).
    https://arxiv.org/abs/1706.03825

    Parameters
    ----------
    model
        The model from which we want to obtain explanations
    output_layer
        Layer to target for the outputs (e.g logits or after softmax).
        If an `int` is provided it will be interpreted as a layer index.
        If a `string` is provided it will look for the layer name.

        Default to the last layer.
        It is recommended to use the layer before Softmax.
    batch_size
        Number of inputs to explain at once, if None compute all at once.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].
    nb_samples
        Number of noisy samples generated for the smoothing procedure.
    noise
        Scalar, noise used as standard deviation of a normal law centered on zero.
    """

    def __init__(self,
                 model: tf.keras.Model,
                 output_layer: Optional[Union[str, int]] = None,
                 batch_size: Optional[int] = 32,
                 operator: Optional[Union[Tasks, str, OperatorSignature]] = None,
                 nb_samples: int = 50,
                 noise: float = 0.2):
        super().__init__(model, output_layer, batch_size, operator)
        self.nb_samples = nb_samples
        self.noise = noise

    @sanitize_input_output
    def explain(self,
                inputs: Union[tf.data.Dataset, tf.Tensor, np.ndarray],
                targets: Optional[Union[tf.Tensor, np.ndarray]] = None) -> tf.Tensor:
        """
        Compute SmoothGrad for a batch of samples.

        Parameters
        ----------
        inputs
            Dataset, Tensor or Array. Input samples to be explained.
            If Dataset, targets should not be provided (included in Dataset).
            Expected shape among (N, W), (N, T, W), (N, H, W, C).
            More information in the documentation.
        targets
            Tensor or Array. One-hot encoding of the model's output from which an explanation
            is desired. One encoding per input and only one output at a time. Therefore,
            the expected shape is (N, output_size).
            More information in the documentation.

        Returns
        -------
        explanations
            Smoothed gradients, same shape as the inputs.
        """
        smoothed_gradients = None
        batch_size = self.batch_size or len(inputs)

        noisy_mask = SmoothGrad._get_noisy_mask((self.nb_samples, *inputs.shape[1:]), self.noise)

        for x_batch, y_batch in batch_tensor((inputs, targets),
                                             max(batch_size // self.nb_samples, 1)):
            noisy_inputs = SmoothGrad._apply_noise(x_batch, noisy_mask)
            repeated_targets = repeat_labels(y_batch, self.nb_samples)
            # compute the gradient of each noisy samples generated
            gradients = self.batch_gradient(self.model, noisy_inputs, repeated_targets, batch_size)
            # group by inputs and compute the average gradient
            gradients = tf.reshape(gradients, (-1, self.nb_samples, *gradients.shape[1:]))
            reduced_gradients = self._reduce_gradients(gradients)

            smoothed_gradients = reduced_gradients if smoothed_gradients is None else tf.concat(
                [smoothed_gradients, reduced_gradients], axis=0)

        return smoothed_gradients

    @staticmethod
    def _get_noisy_mask(shape: Tuple[int, int, int, int],
                        noise: float) -> tf.Tensor:
        """
        Create a random noise mask of the specified shape.

        Parameters
        ----------
        shape
            Desired shape, dimension of one sample.
        noise
            Scalar, noise used as standard deviation of a normal law centered on zero.

        Returns
        -------
        noisy_mask
            Noise mask of the specified shape.
        """
        return tf.random.normal(shape, 0.0, noise, dtype=tf.float32)

    @staticmethod
    @tf.function
    def _apply_noise(inputs: tf.Tensor,
                     noisy_mask: tf.Tensor) -> tf.Tensor:
        """
        Duplicate the samples and apply a noisy mask to each of them.

        Parameters
        ----------
        inputs
            Input samples to be explained.
        noisy_mask
            Mask of random noise to apply on a set of interpolations points. With S the number of
            samples, W & H the sample dimensions and C the number of channels.

        Returns
        -------
        noisy_inputs
            Duplicated inputs with noisy mask applied.
        """
        nb_samples = len(noisy_mask)

        noisy_inputs = tf.repeat(tf.expand_dims(inputs, axis=1), repeats=nb_samples, axis=1)
        noisy_inputs = noisy_inputs + noisy_mask
        noisy_inputs = tf.reshape(noisy_inputs, (-1, *noisy_inputs.shape[2:]))

        return noisy_inputs

    @staticmethod
    @tf.function
    def _reduce_gradients(gradients: tf.Tensor) -> tf.Tensor:
        """
        Average the gradients obtained on each noisy samples.

        Parameters
        ----------
        gradients
            Gradients to reduce the sampling dimension for each inputs.

        Returns
        -------
        reduced_gradients
            Single saliency map for each input.
        """
        return tf.reduce_mean(gradients, axis=1)



import numpy as np
import tensorflow as tf

from xplique.attributions import BoundingBoxesExplainer, Rise
from xplique.attributions.object_detector import SegmentationIouCalculator, BoxIouCalculator, \
    ImageObjectDetectorExplainer, ImageObjectDetectorScoreCalculator, IObjectFormater

from ..utils import almost_equal, generate_model, generate_data


def test_iou_mask():
    """Assert the Mask IoU calculation is ok"""
    dtype = np.float32

    m1 = np.array([
        [10, 20, 30, 40]
    ], dtype=dtype)
    m2 = np.array([
        [15, 20, 30, 40]
    ], dtype=dtype)
    m3 = np.array([
        [0, 20, 10, 40]
    ], dtype=dtype)
    m4 = np.array([
        [0, 0, 100, 100]
    ], dtype=dtype)

    iou_calculator = BoxIouCalculator()

    assert almost_equal(iou_calculator.intersect(m1, m2), 300.0 / 400.0)
    assert almost_equal(iou_calculator.intersect(m1, m3), 0.0)
    assert almost_equal(iou_calculator.intersect(m3, m2), 0.0)
    assert almost_equal(iou_calculator.intersect(m1, m4), 400.0 / 10_000)
    assert almost_equal(iou_calculator.intersect(m2, m4), 300.0 / 10_000)
    assert almost_equal(iou_calculator.intersect(m3, m4), 200.0 / 10_000)


def test_iou_segmentation():
    """Assert the segmentation IoU computation is ok"""

    m1 = np.array([
        [0, 0],
        [1, 1],
    ])[None, :, :]
    m2 = np.array([
        [1, 1],
        [0, 0],
    ])[None, :, :]
    m3 = np.array([
        [1, 0],
        [1, 0],
    ])[None, :, :]

    iou_calculator = SegmentationIouCalculator()

    assert almost_equal(iou_calculator.intersect(m1, m2), 0.0)
    assert almost_equal(iou_calculator.intersect(m1, m3), 1.0/3.0)
    assert almost_equal(iou_calculator.intersect(m3, m2), 1.0/3.0)
    assert almost_equal(iou_calculator.intersect(m1, m1), 1.0)
    assert almost_equal(iou_calculator.intersect(m2, m2), 1.0)


def test_image_object_detector():
    """Assert input shape returned is correct"""
    input_shape = (8, 8, 1)
    nb_labels = 2
    x, y = generate_data(input_shape, nb_labels, nb_labels)
    model = generate_model(input_shape, nb_labels)

    method = Rise(model, nb_samples=10)

    obj_ref = tf.cast([
                [0, 0, 100, 100, 0.9, 1.0, 0.0],
            ], tf.float32)

    class BBoxFormater(IObjectFormater):

        def format_objects(self, predictions):
            if np.array_equal(predictions.numpy(), obj_ref.numpy()):
                return obj_ref[:4], obj_ref[4:5], obj_ref[5:]


            bboxes = tf.cast([
                [0, 10, 20, 30],
                [0, 0, 100, 100],
            ], tf.float32)

            proba = tf.cast([0.9, 0.1], tf.float32)

            classif = tf.cast([[1.0, 0.0], [0.0, 1.0]], tf.float32)

            return bboxes, proba, classif

    formater = BBoxFormater()
    explainer = ImageObjectDetectorExplainer(method, formater, BoxIouCalculator())

    phis = explainer(x, obj_ref)

    assert phis.shape == (1, input_shape[0], input_shape[1])




import numpy as np
import tensorflow as tf
from sklearn.linear_model import SGDClassifier

from xplique.concepts import Cav
from ..utils import generate_data, generate_model


def test_shape():
    """Ensure the output shape is correct"""

    input_shapes = [(8, 8, 3), (8, 8, 1)]
    nb_labels = 2
    nb_samples = 100

    for input_shape in input_shapes:
        x, y = generate_data(input_shape, nb_labels, nb_samples)
        model = generate_model(input_shape, nb_labels)

        for layer in model.layers[1:-1]:
            output_shape = layer.output.shape
            cav = Cav(model, layer.name)(x, x)

            # cav should have the same shape as output, except for the batch size
            assert cav.shape == output_shape[1:]


def test_classifier():
    """For a given simple example, ensure that the differents classifiers works"""
    x = np.random.random((5000, 2))
    y = np.array(x[:, 0]**2 + x[:, 1]**2 > 1.0, np.float32) # y = (x1^2 + x2^2) > 1 (not linearly separable)

    model = tf.keras.Sequential([
        tf.keras.layers.Input(2),
        tf.keras.layers.Lambda(lambda x: x ** 2, name="square_layer"),
        tf.keras.layers.Dense(1, activation='sigmoid')
    ])
    model.compile(loss='binary_crossentropy', optimizer=tf.keras.optimizers.Adam(0.1))
    model.fit(x, y, batch_size=32, epochs=1)

    # the second layer (x -> x^2) make the problem linearly separable
    # the cav vector should be [1, 1]
    positive = []
    negative = []
    for i in range(len(y)):
        if y[i]:
            positive.append(x[i])
        else:
            negative.append(x[i])
    positive = np.array(positive)
    negative = np.array(negative)

    cavs = [
        Cav(model, "square_layer", classifier="SGD")(positive, negative),
        Cav(model, "square_layer", classifier="SVC")(positive, negative),
        Cav(model, "square_layer", classifier=SGDClassifier(alpha=0.1))(positive, negative)
    ]

    # the perfect cav vector should be [0.5, 0.5]
    for cav in cavs:
        assert np.sum(np.abs(cav - [0.5, 0.5])) < 0.1



"""
Module related to ForGrad method enchancement
"""

from functools import lru_cache

import tensorflow as tf
import numpy as np


@lru_cache(maxsize=32)
def _low_pass_real_signal_mask(size : int, bandwith : int) -> tf.Tensor:
    """
    Create a low pass filter mask to be applied on a real signal.
    Since the Discrete Fourier Transform of a real signal is Hermitian-symmetric, only returns the
    fft_length / 2 + 1 unique components of the transform.

    Parameters
    ----------
    size
        Size of the mask.
    bandwith
        Bandwith of the low pass filter. The higher the bandwith, the more frequencies are kept.

    Returns
    -------
    mask
        Low pass filter mask of shape (height, width).
    """
    center = (int(size/2), int(size/2))

    y_grid, x_grid = np.ogrid[:size, :size]
    dist_from_center = np.sqrt((x_grid - center[0])**2 + (y_grid-center[1])**2)

    mask = dist_from_center <= bandwith
    # un-center the mask
    mask = tf.signal.fftshift(mask)
    # keep only the unique components
    mask = tf.cast(mask, tf.float32)[:, :size//2 + 1]

    return mask

def forgrad(explanations : tf.Tensor, sigma: int = 15) -> tf.Tensor:
    """
    ForGRAD is a method that enhances any attributions explanations (particularly useful on
    gradients based attribution method) by eliminating high frequencies in the explanations.

    Ref. Gradient strikes back: How filtering out high frequencies improves explanations (2023).
         https://arxiv.org/pdf/2307.09591.pdf

    Parameters
    ----------
    explanations
        List of explanations to filter. Explanation should be at least 3D (batch, height, width)
        and should have the same height and width.
    sigma
        Bandwith of the low pass filter. The higher the sigma, the more frequencies are kept.
        Sigma should be positive and less than image size.
        Default to paper recommendation, 15 for image size 224.

    Returns
    -------
    filtered_explanations
        Explanations low-pass filtered.
    """
    image_size = explanations.shape[1]

    assert image_size == explanations.shape[2], "Explanations should be square."
    assert len(explanations.shape) > 2, "Explanations should be at least 3D (batch, height, width)."
    assert 0 < sigma <= image_size, "Sigma should be positive and less than image size."

    if len(explanations.shape) == 4:
        explanations = tf.reduce_mean(explanations, -1)
    explanations = tf.abs(explanations)

    spectrums = tf.signal.rfft2d(explanations)

    filter_mask = _low_pass_real_signal_mask(explanations.shape[1], sigma)
    filter_mask = tf.cast(filter_mask, tf.complex64)
    filtered_spectrums = spectrums * filter_mask

    filtered_explanations = tf.signal.irfft2d(filtered_spectrums)
    filtered_explanations = tf.abs(filtered_explanations)

    # resize if odd dimension (cut-off pixel)
    if image_size % 2 == 1:
        filtered_explanations = tf.image.resize(filtered_explanations[:, :, :, None],
                                                (image_size, image_size), method="bicubic")

    return filtered_explanations
